import inspect
from scipy.optimize import least_squares
from scipy.integrate import odeint
from scipy.integrate import solve_ivp
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_is_fitted
from sklearn.metrics import r2_score

import numpy as np
import pandas as pd

from rolldecayestimators.substitute_dynamic_symbols import lambdify
from rolldecayestimators.symbols import *
from rolldecayestimators.measure import fft, fft_omega0

class FitError(Exception):pass

class RollDecay(BaseEstimator):

    # Defining the diff equation for this estimator:
    rhs = -phi_dot_dot/(omega0**2) - 2*zeta/omega0*phi_dot
    roll_diff_equation = sp.Eq(lhs=phi, rhs=rhs)
    acceleration = sp.Eq(lhs=phi, rhs=sp.solve(roll_diff_equation, phi.diff().diff())[0])
    functions = {
        'acceleration':lambdify(acceleration.rhs)
    }

    def __init__(self, ftol=1e-09, maxfev=100000, bounds={}, p0={}, fit_method='derivation', omega_regression=True, omega0=None):
        self.is_fitted_ = False

        self.phi_key = 'phi'  # Roll angle [rad]
        self.phi1d_key = 'phi1d'  # Roll velocity [rad/s]
        self.phi2d_key = 'phi2d'  # Roll acceleration [rad/s2]
        self.y_key = self.phi2d_key
        self.boundaries = bounds
        self.p0 = p0
        self.maxfev=maxfev
        self.ftol=ftol
        self.set_fit_method(fit_method=fit_method)
        self.omega_regression = omega_regression
        self.assert_success = True
        self._omega0 = omega0

    @classmethod
    def load(cls,data:{}, X=None):
        """
        Load data and parameters from an existing fitted estimator

        Parameters
        ----------
        data : dict
            Dict containing data for this estimator such as parameters
        X : pd.DataFrame
            DataFrame containing the measurement that this estimator fits (optional).
        Returns
        -------
        estimator
            Loaded with parameters from data and maybe also a loaded measurement X
        """
        return cls._load(data=data, X=X)

    @classmethod
    def _load(cls,data:{}, X=None):
        """
        Load data and parameters from an existing fitted estimator

        Parameters
        ----------
        data : dict
            Dict containing data for this estimator such as parameters
        X : pd.DataFrame
            DataFrame containing the measurement that this estimator fits (optional).
        Returns
        -------
        estimator
            Loaded with parameters from data and maybe also a loaded measurement X
        """
        estimator = cls()
        estimator.load_data(data=data)
        estimator.load_X(X=X)
        return estimator

    def load_data(self,data:{}):
        parameter_names = self.parameter_names
        missing = list(set(parameter_names) - set(data.keys()))
        if len(missing) > 0:
            raise ValueError('The following parameters are missing in data:%s' % missing)

        parameters = {key: value for key, value in data.items() if key in parameter_names}
        self.parameters = parameters
        self.is_fitted_ = True

    def load_X(self, X=None):
        if isinstance(X, pd.DataFrame):
            self.X=X

    def set_fit_method(self,fit_method):
        self.fit_method = fit_method

        if self.fit_method == 'derivation':
            self.y_key=self.phi2d_key
        elif self.fit_method == 'integration':
            self.y_key=self.phi_key
        else:
            raise ValueError('Unknown fit_mehod:%s' % self.fit_method)

    def __repr__(self):
        if self.is_fitted_:
            parameters = ''.join('%s:%0.3f, '%(key,value) for key,value in sorted(self.parameters.items()))[0:-1]
            return '%s(%s)' % (self.__class__.__name__,parameters)
        else:
            return '%s' % (self.__class__.__name__)

    @property
    def calculate_acceleration(self):
        return self.functions['acceleration']

    @property
    def parameter_names(self):
        signature = inspect.signature(self.calculate_acceleration)

        remove = [self.phi_key, self.phi1d_key]
        if not self.omega_regression:
            remove.append('omega0')

        return list(set(signature.parameters.keys()) - set(remove))

    @staticmethod
    def error(x, self, xs, ys):
        #return np.sum((ys - self.estimator(x, xs))**2)
        return ys-self.estimator(x, xs)


    def estimator(self, x, xs):
        self.parameters = {key: x for key, x in zip(self.parameter_names, x)}

        if not self.omega_regression:
            self.parameters['omega0'] = self.omega0

        if self.fit_method=='derivation':
            self.parameters['phi'] = xs[self.phi_key]
            self.parameters['phi1d'] = xs[self.phi1d_key]
            return self.estimator_acceleration(parameters=self.parameters)
        elif self.fit_method=='integration':
            t = xs.index
            phi0=xs.iloc[0][self.phi_key]
            phi1d0 = xs.iloc[0][self.phi1d_key]

            return self.estimator_integration(t=t, phi0=phi0, phi1d0=phi1d0)
        else:
            raise ValueError('Unknown fit_mehod:%s' % self.fit_method)

    def estimator_acceleration(self,parameters):
        acceleration = self.calculate_acceleration(**parameters)
        return acceleration

    def estimator_integration(self, t, phi0, phi1d0):

        try:
            df = self.simulate(t=t, phi0=phi0, phi1d0=phi1d0)
        except:
            df = pd.DataFrame(index=t)
            df['phi']=np.inf
            df['phi1d']=np.inf

        return df[self.y_key]

    def fit(self, X, y=None, **kwargs):
        self.X = X.copy()

        kwargs = {'self': self,
                  'xs': X,
                  'ys': X[self.y_key]}


        if self.fit_method=='integration':
            self.result = least_squares(fun=self.error, x0=self.initial_guess, kwargs=kwargs, bounds=self.bounds,
                                    ftol=self.ftol, max_nfev=self.maxfev, loss='soft_l1', f_scale=0.1)
        else:
            self.result = least_squares(fun=self.error, x0=self.initial_guess, kwargs=kwargs,
                                    ftol=self.ftol, max_nfev=self.maxfev, method='lm')

        if self.assert_success:
            if not self.result['success']:
                raise FitError(self.result['message'])

        self.parameters = {key: x for key, x in zip(self.parameter_names, self.result.x)}

        if not self.omega_regression:
            self.parameters['omega0'] = self.omega0

        self.is_fitted_ = True

    def simulate(self, t, phi0, phi1d0)->pd.DataFrame:

        states0 = [phi0, phi1d0]

        #states = odeint(self.roll_decay_time_step, y0=states0, t=t, args=(self,parameters))
        #df[self.phi_key] = states[:, 0]
        #df[self.phi1d_key] = states[:, 1]

        t_ = t-t[0]
        t_span = [t_[0], t_[-1]]
        self.simulation_result = solve_ivp(fun=self.roll_decay_time_step, t_span=t_span, y0=states0, t_eval=t_,
                                           )
        if not self.simulation_result['success']:
            raise ValueError('Simulation failed')

        df = pd.DataFrame(index=t)
        df[self.phi_key] = self.simulation_result.y[0, :]
        df[self.phi1d_key] = self.simulation_result.y[1, :]
        p_old = df[self.phi1d_key]
        phi_old = df[self.phi_key]
        df[self.phi2d_key] = self.calculate_acceleration(phi1d=p_old, phi=phi_old, **self.parameters)

        return df

    def roll_decay_time_step(self,t,states):
        # states:
        # [phi,phi1d]

        phi_old = states[0]
        p_old = states[1]

        phi1d = p_old
        calculate_acceleration = self.calculate_acceleration
        phi2d = calculate_acceleration(phi1d=p_old, phi=phi_old, **self.parameters)

        d_states_dt = np.array([phi1d, phi2d])

        return d_states_dt

    def predict(self, X)->pd.DataFrame:

        check_is_fitted(self, 'is_fitted_')

        phi0 = X[self.phi_key].iloc[0]
        phi1d0 = X[self.phi1d_key].iloc[0]
        t = np.array(X.index)
        return self.simulate(t=t, phi0=phi0, phi1d0=phi1d0)


    def score(self, X=None, y=None, sample_weight=None):
        """
        Return the coefficient of determination R_b^2 of the prediction.

        The coefficient R_b^2 is defined as (1 - u/v), where u is the residual sum of squares
        ((y_true - y_pred) ** 2).sum() and v is the total sum of squares ((y_true - y_true.mean()) ** 2).sum().
        The best possible score is 1.0 and it can be negative (because the model can be arbitrarily worse).
        A constant model that always predicts the expected value of y, disregarding the input features,
        would get a R_b^2 score of 0.0.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Test samples. For some estimators this may be a precomputed kernel matrix or a list of generic
            objects instead, shape = (n_samples, n_samples_fitted), where n_samples_fitted is the number of samples
            used in the fitting for the estimator.

        y : Dummy not used

        sample_weight : Dummy

        Returns
        -------
        score : float
            R_b^2 of self.predict(X) wrt. y.

        """

        y_true, y_pred = self.true_and_prediction(X=X)

        return r2_score(y_true=y_true, y_pred=y_pred)

    def true_and_prediction(self, X=None):

        if X is None:
            X=self.X

        y_true = X[self.phi_key]
        df_sim = self.predict(X)
        y_pred = df_sim[self.phi_key]
        return y_true, y_pred

    @property
    def bounds(self):

        minimums = []
        maximums = []

        for key in self.parameter_names:

            boundaries = self.boundaries.get(key,(-np.inf, np.inf))
            assert len(boundaries) == 2
            minimums.append(boundaries[0])
            maximums.append(boundaries[1])

        return [tuple(minimums), tuple(maximums)]

    @property
    def initial_guess(self):
        p0 = []
        for key in self.parameter_names:
            p0.append(self.p0.get(key,0.5))

        return p0

    @property
    def omega0(self):
        """
        Mean natural frequency
        Returns
        -------

        """

        if not self._omega0 is None:
            return self._omega0

        frequencies, dft = fft(self.X['phi'])
        omega0 = fft_omega0(frequencies=frequencies, dft=dft)

        return omega0


    def result_for_database(self, meta_data={}, score=True):
        check_is_fitted(self, 'is_fitted_')

        s = {}
        s.update(self.parameters)
        if score:
            s['score'] = self.score(X=self.X)

        if not self.X is None:
            s['phi_start'] = self.X.iloc[0]['phi']
            s['phi_stop'] = self.X.iloc[-1]['phi']

        if hasattr(self,'omega0'):
            s['omega0_fft'] = self.omega0

        self.results = s  # Store it also

        return s






import numpy as np
import pandas as pd
from scipy.integrate import odeint
from rolldecayestimators import DirectEstimator
from rolldecayestimators.symbols import *
from rolldecayestimators import equations, symbols
from rolldecayestimators.substitute_dynamic_symbols import lambdify, run
from sklearn.utils.validation import check_is_fitted
from rolldecayestimators.estimator import RollDecay



class EstimatorCubic(DirectEstimator):
    """ A template estimator to be used as a reference implementation.

    For more information regarding how to build your own estimator, read more
    in the :ref:`User Guide <user_guide>`.

    Parameters
    ----------
    demo_param : str, default='demo_param'
        A parameter used for demonstation of how to pass and store paramters.
    """

    ## Cubic model:
    b44_cubic_equation = sp.Eq(B_44, B_1 * phi_dot + B_2 * phi_dot * sp.Abs(phi_dot) + B_3 * phi_dot ** 3)
    restoring_equation_cubic = sp.Eq(C_44, C_1 * phi + C_3 * phi ** 3 + C_5 * phi ** 5)

    subs = [
        (B_44, sp.solve(b44_cubic_equation, B_44)[0]),
        (C_44, sp.solve(restoring_equation_cubic, C_44)[0])
    ]
    roll_decay_equation = equations.roll_decay_equation_general_himeno.subs(subs)
    # Normalizing with A_44:
    lhs = (roll_decay_equation.lhs / A_44).subs(equations.subs_normalize).simplify()
    roll_decay_equation_A = sp.Eq(lhs=lhs, rhs=0)

    acceleration = sp.solve(roll_decay_equation_A, phi_dot_dot)[0]
    functions = {
                'acceleration':lambdify(acceleration)
                }

    C_1_equation = equations.C_equation_linear.subs(symbols.C, symbols.C_1)  # C_1 = GM*gm

    eqs = [
        C_1_equation,
        equations.normalize_equations[symbols.C_1]
    ]

    A44_equation = sp.Eq(symbols.A_44, sp.solve(eqs, symbols.C_1, symbols.A_44)[symbols.A_44])
    functions['A44'] = lambdify(sp.solve(A44_equation, symbols.A_44)[0])

    eqs = [equations.C_equation_linear,
           equations.omega0_equation,
           A44_equation,
           ]
    omgea0_equation = sp.Eq(symbols.omega0, sp.solve(eqs, symbols.A_44, symbols.C, symbols.omega0)[0][2])
    functions['omega0'] = lambdify(sp.solve(omgea0_equation,symbols.omega0)[0])

    def __init__(self, maxfev=1000, bounds={}, ftol=10 ** -15, p0={}, fit_method='integration'):

        new_bounds={
            'B_1A':(0, np.inf),  # Assuming only positive coefficients
        #    'B_2A': (0, np.inf),  # Assuming only positive coefficients
        #    'B_3A': (0, np.inf),  # Assuming only positive coefficients
        }
        new_bounds.update(bounds)
        bounds=new_bounds

        super().__init__(maxfev=maxfev, bounds=bounds, ftol=ftol, p0=p0, fit_method=fit_method, omega_regression=True)


    @classmethod
    def load(cls, B_1A:float, B_2A:float, B_3A:float, C_1A:float, C_3A:float, C_5A:float, X=None, **kwargs):
        """
        Load data and parameters from an existing fitted estimator

        A_44 is total roll intertia [kg*m**2] (including added mass)

        Parameters
        ----------
        B_1A
            B_1/A_44 : linear damping
        B_2A
            B_2/A_44 : quadratic damping
        B_3A
            B_3/A_44 : cubic damping
        C_1A
            C_1/A_44 : linear stiffness
        C_3A
            C_3/A_44 : cubic stiffness
        C_5A
            C_5/A_44 : pentatonic stiffness

        X : pd.DataFrame
            DataFrame containing the measurement that this estimator fits (optional).
        Returns
        -------
        estimator
            Loaded with parameters from data and maybe also a loaded measurement X
        """
        data={
            'B_1A':B_1A,
            'B_2A':B_2A,
            'B_3A':B_3A,
            'C_1A':C_1A,
            'C_3A':C_3A,
            'C_5A':C_5A,
        }

        return super(cls, cls)._load(data=data, X=X)


    def calculate_additional_parameters(self, A44):
        check_is_fitted(self, 'is_fitted_')

        parameters_additional = {}

        for key, value in self.parameters.items():
            symbol_key = sp.Symbol(key)
            new_key = key[0:-1]
            symbol_new_key = ss.Symbol(new_key)

            if symbol_new_key in equations.normalize_equations:
                normalize_equation = equations.normalize_equations[symbol_new_key]
                solution = sp.solve(normalize_equation,symbol_new_key)[0]
                new_value = solution.subs([(symbol_key,value),
                                           (symbols.A_44,A44),

                               ])

                parameters_additional[new_key]=new_value

        return parameters_additional


    def result_for_database(self, meta_data={}):
        s = super().result_for_database(meta_data=meta_data)

        inputs=pd.Series(meta_data)

        inputs['m'] = inputs['Volume']*inputs['rho']
        parameters = pd.Series(self.parameters)
        inputs = parameters.combine_first(inputs)

        s['A_44'] = run(self.functions['A44'], inputs=inputs)
        parameters_additional = self.calculate_additional_parameters(A44=s['A_44'])
        s.update(parameters_additional)

        inputs['A_44'] = s['A_44']
        s['omega0'] = run(function=self.functions['omega0'], inputs=inputs)

        self.results = s  # Store it also

        return s


class EstimatorQuadraticB(EstimatorCubic):
    """ A template estimator to be used as a reference implementation.

    For more information regarding how to build your own estimator, read more
    in the :ref:`User Guide <user_guide>`.

    Parameters
    ----------
    demo_param : str, default='demo_param'
        A parameter used for demonstation of how to pass and store paramters.
    """

    ## Cubic model:
    b44_quadratic_equation = sp.Eq(B_44, B_1 * phi_dot + B_2 * phi_dot * sp.Abs(phi_dot))
    restoring_equation_quadratic = sp.Eq(C_44, C_1 * phi)

    subs = [
        (B_44, sp.solve(b44_quadratic_equation, B_44)[0]),
        (C_44, sp.solve(restoring_equation_quadratic, C_44)[0])
    ]
    roll_decay_equation = equations.roll_decay_equation_general_himeno.subs(subs)
    # Normalizing with A_44:
    lhs = (roll_decay_equation.lhs / A_44).subs(equations.subs_normalize).simplify()
    roll_decay_equation_A = sp.Eq(lhs=lhs, rhs=0)

    acceleration = sp.solve(roll_decay_equation_A, phi_dot_dot)[0]
    functions = dict(EstimatorCubic.functions)
    functions['acceleration'] = lambdify(acceleration)

    @classmethod
    def load(cls, B_1A:float, B_2A:float, C_1A:float, X=None, **kwargs):
        """
        Load data and parameters from an existing fitted estimator

        A_44 is total roll intertia [kg*m**2] (including added mass)

        Parameters
        ----------
        B_1A
            B_1/A_44 : linear damping
        B_2A
            B_2/A_44 : quadratic damping
        C_1A
            C_1/A_44 : linear stiffness

        X : pd.DataFrame
            DataFrame containing the measurement that this estimator fits (optional).
        Returns
        -------
        estimator
            Loaded with parameters from data and maybe also a loaded measurement X
        """
        data={
            'B_1A':B_1A,
            'B_2A':B_2A,
            'C_1A':C_1A,
        }

        return super(cls, cls)._load(data=data, X=X)


class EstimatorQuadraticBandC(EstimatorCubic):
    """ A template estimator to be used as a reference implementation.

    For more information regarding how to build your own estimator, read more
    in the :ref:`User Guide <user_guide>`.

    Parameters
    ----------
    demo_param : str, default='demo_param'
        A parameter used for demonstation of how to pass and store paramters.
    """

    ## Quadratic model:
    b44_quadratic_equation = sp.Eq(B_44, B_1 * phi_dot + B_2 * phi_dot * sp.Abs(phi_dot))
    restoring_equation_quadratic = sp.Eq(C_44, C_1 * phi + C_3 * phi ** 3)

    subs = [
        (B_44, sp.solve(b44_quadratic_equation, B_44)[0]),
        (C_44, sp.solve(restoring_equation_quadratic, C_44)[0])
    ]
    roll_decay_equation = equations.roll_decay_equation_general_himeno.subs(subs)
    # Normalizing with A_44:
    lhs = (roll_decay_equation.lhs / A_44).subs(equations.subs_normalize).simplify()
    roll_decay_equation_A = sp.Eq(lhs=lhs, rhs=0)

    acceleration = sp.solve(roll_decay_equation_A, phi_dot_dot)[0]
    functions = dict(EstimatorCubic.functions)
    functions['acceleration'] = lambdify(acceleration)

class EstimatorLinear(EstimatorCubic):
    """ A template estimator to be used as a reference implementation.

    For more information regarding how to build your own estimator, read more
    in the :ref:`User Guide <user_guide>`.

    Parameters
    ----------
    demo_param : str, default='demo_param'
        A parameter used for demonstation of how to pass and store paramters.
    """

    ## Linear model:
    b44_linear_equation = sp.Eq(B_44, B_1 * phi_dot)
    restoring_linear_quadratic = sp.Eq(C_44, C_1 * phi)


    subs = [
        (B_44, sp.solve(b44_linear_equation, B_44)[0]),
        (C_44, sp.solve(restoring_linear_quadratic, C_44)[0])
    ]
    roll_decay_equation = equations.roll_decay_equation_general_himeno.subs(subs)
    # Normalizing with A_44:
    lhs = (roll_decay_equation.lhs / A_44).subs(equations.subs_normalize).simplify()
    roll_decay_equation_A = sp.Eq(lhs=lhs, rhs=0)

    acceleration = sp.solve(roll_decay_equation_A, phi_dot_dot)[0]
    functions = dict(EstimatorCubic.functions)
    functions['acceleration'] = lambdify(acceleration)

    @classmethod
    def load(cls, B_1A:float, C_1A:float, X=None, **kwargs):
        """
        Load data and parameters from an existing fitted estimator

        A_44 is total roll intertia [kg*m**2] (including added mass)

        Parameters
        ----------
        B_1A
            B_1/A_44 : linear damping
        C_1A
            C_1/A_44 : linear stiffness

        X : pd.DataFrame
            DataFrame containing the measurement that this estimator fits (optional).
        Returns
        -------
        estimator
            Loaded with parameters from data and maybe also a loaded measurement X
        """
        data={
            'B_1A':B_1A,
            'C_1A':C_1A,
        }

        return super(cls, cls)._load(data=data, X=X)





import pytest
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from rolldecayestimators.measure import fft,fft_omega0
from numpy.testing import assert_almost_equal

omega0=2.0
T0 = 2*np.pi/omega0
A = 1.0

@pytest.fixture
def signal():
    t = np.arange(0,10.5*T0,0.01)
    y = A*np.sin(omega0*t)
    s = pd.Series(data=y, index=t)
    yield s

@pytest.fixture
def signal2():
    t = np.arange(0,5.0*T0,0.01)
    y = A*np.sin(omega0*t)
    s = pd.Series(data=y, index=t)
    yield s

def test_fft(signal):

    fig, ax = plt.subplots()
    signal.plot(ax=ax)
    plt.show()

    frequencies, dft = fft(signal)

    index = np.argmax(dft)
    natural_frequency = frequencies[index]
    dft_predict = dft[index]
    omega0_predict = 2 * np.pi * natural_frequency


    fig,ax = plt.subplots()
    ax.plot(2*np.pi*frequencies,dft, '.-', label='dft')
    ax.plot(omega0_predict, dft_predict, 'ro', label='omega_0=%f'%omega0_predict)
    ax.set_xlim(0,omega0_predict+5)
    ax.legend()

    plt.show()

#@pytest.mark.skip('Fix this one...')
def test_fft_omega0(signal):

    frequencies, dft = fft(signal)
    omega0_predict = fft_omega0(frequencies=frequencies, dft=dft)
    assert_almost_equal(omega0_predict, omega0, decimal=2)

def test_fft_omega0_even(signal2):

    frequencies, dft = fft(signal2)
    omega0_predict = fft_omega0(frequencies=frequencies, dft=dft)
    assert_almost_equal(omega0_predict, omega0, decimal=2)

def signal_generator(t,omega0):
    y = A*np.sin(omega0*t)
    s = pd.Series(data=y, index=t)
    return s

def test_fft_omega0_even_many():

    omega0s = np.linspace(0.1,6,10)

    for omega0_ in omega0s:
        T0_ = 2 * np.pi / omega0_
        t = np.arange(0, 10.5*T0_, 0.01)

        s = signal_generator(t=t, omega0=omega0_)

        frequencies, dft = fft(s)
        omega0_predict = fft_omega0(frequencies=frequencies, dft=dft)
        assert_almost_equal(omega0_predict, omega0_, decimal=2)

#@pytest.mark.skip('Fix this one...')
def test_fft_omega0_many():

    omega0s = np.linspace(0.1,6,10)

    for omega0_ in omega0s:
        T0_ = 2 * np.pi / omega0_
        t = np.arange(0, 10.5*T0_, 0.01)

        s = signal_generator(t=t, omega0=omega0_)

        frequencies, dft = fft(s)
        omega0_predict = fft_omega0(frequencies=frequencies, dft=dft)
        assert_almost_equal(omega0_predict, omega0_, decimal=2)





"""
This is a translation of Carl-Johans implementation in Matlab to Python
"""
import os.path
import numpy as np
from numpy import tanh, exp, sqrt, pi, sin, cos, arctan
import pandas as pd
import scipy.interpolate
import matplotlib.pyplot as plt

dir_path = os.path.dirname(__file__)
base_path = os.path.split(dir_path)[0]

data_path_S175 = os.path.join(base_path, 'rolldecayestimators', 'Bw0_S175.csv')
data_S175 = pd.read_csv(data_path_S175, sep=';')
f_interpolation_S175 = scipy.interpolate.interp1d(data_S175['w_vec'], data_S175['b44_vec'], kind='cubic')

data_path_faust = os.path.join(base_path, 'rolldecayestimators', 'Bw0_faust.csv')
data_faust = pd.read_csv(data_path_faust, sep=';')
f_interpolation_faust = scipy.interpolate.interp1d(data_faust['w_vec'], data_faust['b44_vec'], kind='cubic')

def Bw0_S175(w):
    Bw0 = f_interpolation_S175(w)
    return Bw0

def Bw0_faust(w):
    Bw0 = f_interpolation_faust(w)
    return Bw0

def Bw_S175(w, V, d, g=9.81):
    Bw0 = Bw0_S175(w)
    BW44 = Bw(w, V, d, Bw0, g=9.81)

    return BW44

def Bw_faust(w, V, d, g=9.81):
    Bw0 = Bw0_faust(w)
    BW44 = Bw(w, V, d, Bw0, g=9.81)

    return BW44

def Bw(w, V, d, Bw0, g=9.81):
    OMEGA = w * V / g
    zeta_d = w ** 2 * d / g
    A1 = 1 + zeta_d ** (-1.2) * exp(-2 * zeta_d)
    A2 = 0.5 + zeta_d ** (-1) * exp(-2 * zeta_d)

    Bw_div_Bw0 = 0.5 * (
            ((A1 + 1) + (A2 - 1) * tanh(20 * (OMEGA - 0.3))) + (2 * A1 - A2 - 1) * exp(-150 * (OMEGA - 0.25) ** 2))

    BW44 = Bw0 * Bw_div_Bw0
    return BW44


def bilge_keel(w, fi_a, V, B, d, A, bBK, R, g, OG, Ho, ra):
    """
    ITTC
    definitions

    B44BK = B44BK_N0 + B44BK_H0 + B44BK_L + B44BK_W
    """

    tata = A / (B * d);


    # Normal Force component(ITTC)
    f = 1 + 0.3 * exp(-160 * (1 - tata));
    l = d * sqrt((Ho - (1 - sqrt(2) / 2) * R / d) ** 2 + (1 - OG / d - (1 - sqrt(2) / 2) * R / d) ** 2) # distance from CoG to tip of bilge keel
    CD = 22.5 * bBK / (pi * l * fi_a * f) + 2.4;
    Bp44BK_N0 = 8 / (3 * pi) * ra * l ** 3 * w * fi_a * bBK * f * CD;

    # #
    # Hull pressure component
    So = 0.3 * pi * l * fi_a * f / bBK + 1.95; # S175 implementation
    m1 = R / d;
    m2 = OG / d;
    m3 = 1 - m1 - m2;
    m4 = Ho - m1;
    m5 = (0.414 * Ho + 0.0651 * m1 ** 2 - (0.382 * Ho + 0.0106) * m1) / ((Ho - 0.215 * m1) * (1 - 0.215 * m1));
    m6 = (0.414 * Ho + 0.0651 * m1 ** 2 - (0.382 + 0.0106 * Ho) * m1) / ((Ho - 0.215 * m1) * (1 - 0.215 * m1));

    if So > 0.25 * pi * R:
        m7 = So / d - 0.25 * pi * m1;
    else:
        m7 = 0;

    if So > 0.25 * pi * R:
        m8 = m7 + 0.414 * m1;
    else:
        m8 = m7 + 1.414 * m1 * (1 - cos(So / R));

    Ao = (m3 + m4) * m8 - m7 ** 3;
    Bo = m2 ** 2 / (3 * (Ho - 0.215 / m1)) + (1 - m1) ** 2 * (2 * m3 - m2) / (6 * (1 - 0.215 * m1)) + m1 * (
                m3 * m5 + m4 * m6);
    Cp_minus = -22.5 * bBK / (pi * l * fi_a * f) - 1.2;
    Cp_plus = 1.2;

    Bp44BK_H0 = 4 / (3 * pi) * ra * l ** 2 * w * fi_a * d ** 2 * (-Ao * Cp_minus + Bo * Cp_plus);

    # # Ikeda 1994 bilge keel generated Lift # Fartberoende??
    l1 = l + bBK / 2; # Lift Force
    u = l1 * fi_a * w; # tangential velocity
    alpha = arctan(u / V); # flow velocity?
    Vr = sqrt(V ** 2 + u ** 2); # -'' -

    LBK = pi * ra * alpha * Vr ** 2 * bBK ** 2 / 2; # lift

    B44BK_L = 2 * LBK * l1 / (fi_a * w); #

    # #
    # wave making contribution from bilge keels, normally very small...,
    # non - dimensional!!!!!

    C_BK = bBK / B; # estimation according to ITTC
    lBK = l - bBK; # needs to be verified!!!!
    fi = fi_a;

    dBK = lBK * ((2 * d / B) / sqrt(1 + (2 * d / B) ** 2) * cos(fi) - sin(fi) / (1 + (2 * d / B) ** 2));  # S175 implementation

    B44BKW0 = C_BK * exp(-w ** 2 / g * dBK); # non dimensional  wave damping from BK, ITTC
    return Bp44BK_N0, Bp44BK_H0, B44BK_L, B44BKW0

def frictional(w, fi_a, V, B, d, OG, ra, Cb, L, visc =1.15 * 10 ** -6):
    # ITTC

    Sf   = L*(1.7*d+Cb*B); # Wetted surface approx
    r_f  = 1/pi*((0.887+0.145*Cb)*(Sf/L)+2*OG);  # S175


    Rn = 0.512 * (r_f/fi_a) ** 2 * w / visc;
    Cf = 1.328*Rn**-0.5+0.14*Rn**-0.114;
    B44F0=0.5*ra*r_f**3*Sf*Cf;
    B44F= B44F0 * 8 / (3*pi) * fi_a * w * (1 + 4.1 * V / (w * L))
    return B44F

def hull_lift(V,B, d, OG, ra, L, A):

    lo   = 0.3*d;
    lR   = 0.5*d;
    K    = 0.1;         #!! depends on CM  # S175
    #C_mid = A/(B*d)
    #K = 106 * (C_mid - 0.91)**2 - 700 * (C_mid - 0.91)**3;

    kN   = 2*pi*d/L+K*(4.1*B/L-0.045);
    B44L = ra/2*V*L*d*kN*lo*lR*(1+1.4*OG/lR+0.7*OG**2/(lo*lR))  # S175

    return B44L

def calculate_B44(w, V, d, Bw0, fi_a,  B,  A, bBK, R, OG, Ho, ra, Cb, L, LBK, visc =   1.15*10**-6, g=9.81):
    BW44=Bw(w, V, d, Bw0, g=9.81)

    Bp44BK_N0, Bp44BK_H0, B44BK_L, B44BKW0 = bilge_keel(w, fi_a, V, B, d, A, bBK, R, g, OG, Ho, ra)
    B44BK_N0 = Bp44BK_N0*LBK
    B44BK_H0 = Bp44BK_H0*LBK
    B44BK_L = B44BK_L
    # B44BKW0 = B44BKW0 * dim...
    B44_BK = B44BK_N0 + B44BK_H0 + B44BK_L

    B44F = frictional(w,fi_a,V,B, d, OG, ra, Cb, L, visc)

    B44L = hull_lift(V,B, d, OG, ra, L, A)

    B44 = BW44+B44_BK+B44F+B44L

    return B44,BW44,B44_BK,B44F,B44L

def calculate_B44_series(row):
    s = pd.Series(name=row.name)
    B_44,B_W,B_BK,B_F,B_L = calculate_B44(w=row['w'], V=row['V'], d=row['d'], Bw0=row['Bw0'], fi_a=row['fi_a'], B=row['B'], A=row['A'], bBK=row['BKB'], R=row['R_b'],
                         OG=row['OG'], Ho=row['Ho'], ra=row['rho'], Cb=row['Cb'], L=row['L'], LBK=row['LBK'], visc=row['visc'], g=row['g'])
    s['B_44']=B_44
    s['B_W']=B_W
    s['B_BK']=B_BK
    s['B_F']=B_F
    s['B_L']=B_L
    return s














import pytest
import pandas as pd
import os.path
from sklearn.datasets import make_regression
from sklearn.pipeline import Pipeline
from sklearn.feature_selection import SelectKBest
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import PolynomialFeatures
from sklearn.feature_selection import f_regression
from sklearn.feature_selection import VarianceThreshold
import sympy as sp

from rolldecayestimators.polynom_estimator import Polynom
import matplotlib.pyplot as plt

@pytest.fixture
def data():
    # Generate some random regression data:
    X, y = make_regression(n_samples=1000, n_features=1, n_informative=10, noise=10, random_state=1)
    yield X, y


@pytest.fixture
def model(data):

    X=data[0]
    y=data[1]

    polynomial_features = PolynomialFeatures(degree=1)
    variance_treshold = VarianceThreshold()
    linear_regression = LinearRegression()

    select_k_best = SelectKBest(k=1, score_func=f_regression)
    steps = [
        ('polynomial_feature', polynomial_features),
        # ('standard_scaler', standard_scaler),
        ('variance_treshold', variance_treshold),
        ('select_k_best', select_k_best),
        ('linear_regression', linear_regression)
    ]

    model = Pipeline(steps=steps)
    model.fit(X=X, y=y)

    yield model


def test_fit(data, model):

    X=data[0]
    x=X[:,0]
    y=data[1]

    y_symbol = sp.symbols('y')
    polynom = Polynom(model=model, columns=['B_e_hat'], y_symbol=y_symbol)
    polynom.fit(X=X, y=y)

    fig,ax=plt.subplots()
    ax.plot(x,y, '.')
    ax.plot(x, model.predict(X=X), label='model')
    ax.plot(x, polynom.predict(X=X), '--', label='polynom')

    ax.legend()
    plt.show()

@pytest.fixture
def polynom(data, model):

    X = data[0]
    y = data[1]
    y_symbol = sp.symbols('y')
    polynom = Polynom(model=model, columns=['B_e_hat'], y_symbol=y_symbol)
    polynom.fit(X=X, y=y)
    yield polynom

def test_predict_dict(polynom):

    data = {
        'B_e_hat':1,
    }

    polynom.predict(X=data)

def test_predict_series(polynom):

    data = {
        'B_e_hat':1,
    }
    data = pd.Series(data)

    polynom.predict(X=data)

def test_save_load(tmpdir, polynom):

    file_path=os.path.join(str(tmpdir), 'test.sym')
    polynom.save(file_path=file_path)
    polynom2 = Polynom.load(file_path=file_path)
    data = {
        'B_e_hat': 1,
    }

    polynom2.predict(X=data)




"""
Python conversion of fortran code by Martin Alexandersson
************************************************************************
*  Simple Prediction Formula of Roll Damping                           *
*                         on the Basis of Ikeda's Method               *
*                                                                      *
*  Roll_Damping.for            coded by Yoshiho IKEDA                  *
*                                       Yuki KAWAHARA                  *
*                                       Kazuya MAEKAWA                 *
*                                       Osaka Prefecture University    *
*                                       Graduate School of Engineering *
*  last up date 2009.07.08                                             *
************************************************************************
"""

from numpy import exp as EXP
from numpy import log as LOG
from numpy import sqrt as SQRT
from numpy import pi as PI
import numpy as np
import pandas as pd
from rolldecayestimators import ikeda_speed

def calculate_roll_damping(LPP, Beam, CB, CMID, OG, PHI, BKL, BKB, OMEGA,
                           DRAFT, V=0, KVC = 1.14e-6, verify_input=True, limit_inputs=False, Bw_div_Bw0_max=np.inf,
                           BWHAT_lim=np.inf, rho=1000, alternative_bilge_keel=False, RdivB=0.02):
    """
    ********************************************************************
    *** Calculation of roll damping by the proposed predition method ***
    ********************************************************************

    :param LPP: [m]
    :param Beam: [m]
    :param CB: Block coefficient [-]
    :param CMID: Middship coefficient (A_m/(B*d) [-]
    :param OG: distance from the still water level O to the roll axis G [m]
    :param PHI: Roll angle [deg]
    :param BKL: length of bilge keel [m]
    :param BKB: height of bilge keel [m]
    :param OMEGA: Frequency of motion [rad/s]
    :param DRAFT: DRAFT : ship draught [m]
    :param OMEGAHAT:
    :param V: ship speed [m/s]
    :param KVC = 1.14e-6  # Kinematic Viscosity Coefficient
    :param verify_input = True, should the inputs be verified to be within limits?
    :param limit_inputs = False, use limit value if input limit is exceeded.
    :param Bw_div_Bw0_max=12, There are some problems with the wave damping speed dependence being over predicted this
        is a limit. Set it to np.inf to turn it off.
    :param BWHAT_lim=0.005, There are some problems with the wave damping being over predicted this
        is a limit. Set it to np.inf to turn it off.
    :param rho, water density
    :param alternative_bilge_keel, if True an alternative bilge keel calculation is used.
    :RdivB, bilge radius / ship beam (only used if alternative_bilge_keel=True)

    :return: B44HAT, BFHAT, BWHAT, BEHAT, BBKHAT
     Nondimensional damping:
    B44HAT: Total
    BFHAT: Friction
    BWHAT: Wave
    BEHAT: Eddy
    BBKHAT: Bilge keel
    """

    # Check if any input is null:
    inputs = {
        'LPP': LPP,
        'Beam': Beam,
        'CB': CB,
        'CMID': CMID,
        'OG': OG,
        'PHI': PHI,
        'BKL': BKL,
        'BKB': BKB,
        'OMEGA': OMEGA,
        'DRAFT': DRAFT,
    }
    inputs=pd.Series(inputs)
    mask = pd.isnull(inputs)
    nulls = inputs[mask]
    if len(nulls)>0:
        raise SimplifiedIkedaInputError('%s is NaN' % nulls)

    if limit_inputs:
        # Limit the inputs to not exceed the limits.
        outputs = _limit_inputs(LPP=LPP, Beam=Beam, CB=CB, CMID=CMID, OG=OG, PHI=PHI, lBK=BKL, bBK=BKB, OMEGA=OMEGA, DRAFT=DRAFT)
        LPP = outputs['LPP']
        Beam = outputs['Beam']
        CB = outputs['CB']
        CMID = outputs['CMID']
        OG = outputs['OG']
        PHI = outputs['PHI']
        BKL = outputs['BKL']
        BKB = outputs['BKB']
        OMEGA = outputs['OMEGA']
        DRAFT = outputs['DRAFT']

    if verify_input:
        verify_inputs(LPP, Beam, CB, CMID, OG, PHI, BKL, BKB, OMEGA,
                      DRAFT)

    LBKL= BKL / LPP
    BD = Beam/DRAFT
    OGD = OG/DRAFT
    BBKB = BKB / Beam
    BRTH = Beam
    #OMEGA/= 2  # Magic factor!??? This value seemt o give better results..?
    OMEGAHAT = OMEGA * SQRT(BRTH / 2 / 9.81)

    B44HAT, BFHAT, BWHAT, BEHAT, BBKHAT = _calculate_roll_damping(LPP=LPP,BRTH=BRTH, CB=CB, CMID=CMID, OGD=OGD, PHI=PHI, LBKL=LBKL, BBKB=BBKB,
                                   OMEGA=OMEGA, DRAFT=DRAFT, BD=BD, OMEGAHAT=OMEGAHAT, KVC=KVC)

    # Speed dependance:
    # Hull lift:
    A=CMID*Beam*DRAFT
    BL=ikeda_speed.hull_lift(V=V,B=Beam, d=DRAFT, OG=OG, L=LPP, A=A, ra=1025)
    disp=LPP*Beam*DRAFT*CB
    ND_factorB = np.sqrt(Beam / (2 * 9.81)) / (1025 * disp * (Beam**2));  # Nondimensiolizing factor of B44
    BLHAT=BL*ND_factorB
    B44HAT+=BLHAT

    # Wave speed dependance:
    if V>0:
        BWHAT_speed =ikeda_speed.Bw(w=OMEGA, V=V, d=DRAFT, Bw0=BWHAT, Bw_div_Bw0_max=Bw_div_Bw0_max)

        if BWHAT_speed>BWHAT_lim:
            BWHAT_speed=BWHAT_lim # This limit give some improvement

        B44HAT = B44HAT - BWHAT + BWHAT_speed
        BWHAT = BWHAT_speed

    # Eddy speed dependence:
    if V>0:
        factor=(0.04*OMEGA*LPP/V)**2
        B44HAT-=BEHAT
        BEHAT*=(factor)/(1+factor)
        B44HAT+=BEHAT

    if alternative_bilge_keel and BKB>0:

        B44HAT-=BBKHAT

        A=CMID*Beam*DRAFT
        g=9.81
        Ho = Beam / (2 * DRAFT)  # half breadth to draft ratio
        R = RdivB*Beam
        Bp44BK_N0, Bp44BK_H0, B44BK_L, B44BKW0 = ikeda_speed.bilge_keel(w=OMEGA, fi_a=np.deg2rad(PHI), V=V, B=Beam, d=DRAFT, A=A, bBK=BKB, R=R, g=g, OG=OG, Ho=Ho, ra=rho)
        B44BK_N0 = Bp44BK_N0 * BKL
        B44BK_H0 = Bp44BK_H0 * BKL
        B44BK_L = B44BK_L
        # B44BKW0 = B44BKW0 * dim...
        B44_BK = B44BK_N0 + B44BK_H0 + B44BK_L
        BBKHAT=ND_factorB*B44_BK
        B44HAT+=BBKHAT

    return B44HAT, BFHAT, BWHAT, BEHAT, BBKHAT, BLHAT

def _calculate_roll_damping(LPP, BRTH, CB, CMID, OGD, PHI, LBKL, BBKB, OMEGA,
                           DRAFT, BD, OMEGAHAT, KVC = 1.14e-6):
    """
    ********************************************************************
    *** Calculation of roll damping by the proposed predition method ***
    ********************************************************************

    :param LPP: [m]
    :param Beam: [m]
    :param CB: Block coefficient [-]
    :param CMID: Middship coefficient (A_m/(B*d) [-]
    :param OGD: OG/DRAFT
    :param PHI: Roll angle [deg]
    :param LBKL: BKL/LPP
    :param BBKB : BKB/Beam
    :param OMEGA: Frequency of motion [rad/s]
    :param DRAFT: DRAFT : ship draught [m]
    :param BD: Beam/DRAFT
    :param OMEGAHAT:
    :param KVC = 1.14e-6  # Kinematic Viscosity Coefficient

    :return: B44HAT, BFHAT, BWHAT, BEHAT, BBKHAT
     Nondimensional damping:
    B44HAT: Total
    BFHAT: Friction
    BWHAT: Wave
    BEHAT: Eddy
    BBKHAT: Bilge keel
    """

    # *** Frictional Component ***
    BFHAT = calculate_B_F(BD, BRTH, CB, DRAFT, KVC, LPP, OGD, OMEGA, PHI)

    #*** Wave Component ***
    BWHAT = calculate_B_W0(BD, CB, CMID, OGD, OMEGAHAT)

    #*** Eddy Component ***
    BEHAT = calculate_B_E(BD, CB, CMID, OGD, OMEGAHAT, PHI)

    #*** Bilge Keel Component ***
    BBKHAT = calculate_B_BK(BBKB, BD, CB, CMID, LBKL, OGD, OMEGAHAT, PHI)

    #*** Total Roll Damping ***
    B44HAT=BFHAT+BWHAT+BEHAT+BBKHAT

    return B44HAT, BFHAT, BWHAT, BEHAT, BBKHAT


def calculate_B_BK(BBKB, BD, CB, CMID, LBKL, OGD, OMEGAHAT, PHI):
    """
    Bilge keel damping

    :param BBKB : BKB/Beam
    :param BD: Beam/DRAFT
    :param CB: Block coefficient [-]
    :param CMID: Middship coefficient (A_m/(B*d) [-]
    :param LBKL: BKL/LPP
    :param OGD: OG/DRAFT
    :param OMEGAHAT:
    :param PHI: Roll angle [deg]
    :return: BBKHAT
    """


    FBK1 = (-0.3651 * CB + 0.3907) * (BD - 2.83) ** 2 - 2.21 * CB + 2.632
    FBK2 = 0.00255 * PHI ** 2 + 0.122 * PHI + 0.4794
    FBK3 = (-0.8913 * BBKB ** 2 - 0.0733 * BBKB) * LBKL ** 2 + (
                5.2857 * BBKB ** 2 - 0.01185 * BBKB + 0.00189) * LBKL
    ABK = FBK1 * FBK2 * FBK3
    BBK1 = (5.0 * BBKB + 0.3 * BD - 0.2 * LBKL + 0.00125 * PHI ** 2 - 0.0425 * PHI - 1.86) * OGD
    BBK2 = -15.0 * BBKB + 1.2 * CB - 0.1 * BD - 0.0657 * OGD ** 2 + 0.0586 * OGD + 1.6164
    BBK3 = 2.5 * OGD + 15.75
    BBKHAT = ABK * EXP(BBK1 + BBK2 * CMID ** BBK3) * OMEGAHAT

    LBKL = np.array(LBKL)
    BBKHAT = np.array(BBKHAT)

    mask = LBKL == 0
    BBKHAT[mask]=0

    return BBKHAT


def calculate_B_E(BD, CB, CMID, OGD, OMEGAHAT, PHI):
    """
    Eddy damping B_E

    :param BD: Beam/DRAFT
    :param CB: Block coefficient [-]
    :param CMID: Middship coefficient (A_m/(B*d) [-]
    :param OGD: OG/DRAFT
    :param OMEGAHAT:
    :param PHI: Roll angle [deg]

    :return: BEHAT
    """

    FE1 = (-0.0182 * CB + 0.0155) * (BD - 1.8) ** 3
    FE2 = -79.414 * CB ** 4 + 215.695 * CB ** 3 - 215.883 * CB ** 2 + 93.894 * CB - 14.848
    AE = FE1 + FE2
    BE1 = (3.98 * CB - 5.1525) * (-0.2 * BD + 1.6) * OGD * (
                (0.9717 * CB ** 2 - 1.55 * CB + 0.723) * OGD + 0.04567 * CB + 0.9408)
    BE2 = (0.25 * OGD + 0.95) * OGD - 219.2 * CB ** 3 + 443.7 * CB ** 2 - 283.3 * CB + 59.6
    BE3 = -15 * CB * BD + 46.5 * CB + 11.2 * BD - 28.6
    CR = AE * EXP(BE1 + BE2 * CMID ** BE3)
    BEHAT = 4.0 * OMEGAHAT * PHI * PI / 180 / (3.0 * PI * CB * BD ** 3.0) * CR
    return BEHAT


def calculate_B_W0(BD, CB, CMID, OGD, OMEGAHAT):
    """
    Wave roll damping at zero speed B_W0

    :param BD: Beam/DRAFT
    :param CB: Block coefficient [-]
    :param CMID: Middship coefficient (A_m/(B*d) [-]
    :param OGD: OG/DRAFT
    :param OMEGAHAT:

    :return:B_W0

    """

    X1 = BD;
    X2 = CB;
    X3 = CMID
    X5 = OMEGAHAT
    X4 = 1 - OGD
    A111 = -0.002222 * X1 ** 3 + 0.040871 * X1 ** 2 - 0.286866 * X1 + 0.599424
    A112 = 0.010185 * X1 ** 3 - 0.161176 * X1 ** 2 + 0.904989 * X1 - 1.641389
    A113 = -0.015422 * X1 ** 3 + 0.220371 * X1 ** 2 - 1.084987 * X1 + 1.834167
    A121 = -0.0628667 * X1 ** 4 + 0.4989259 * X1 ** 3 + 0.52735 * X1 ** 2 - 10.7918672 * X1 + 16.616327
    A122 = 0.1140667 * X1 ** 4 - 0.8108963 * X1 ** 3 - 2.2186833 * X1 ** 2 + 25.1269741 * X1 - 37.7729778
    A123 = -0.0589333 * X1 ** 4 + 0.2639704 * X1 ** 3 + 3.1949667 * X1 ** 2 - 21.8126569 * X1 + 31.4113508
    A124 = 0.0107667 * X1 ** 4 + 0.0018704 * X1 ** 3 - 1.2494083 * X1 ** 2 + 6.9427931 * X1 - 10.2018992
    A131 = 0.192207 * X1 ** 3 - 2.787462 * X1 ** 2 + 12.507855 * X1 - 14.764856
    A132 = -0.350563 * X1 ** 3 + 5.222348 * X1 ** 2 - 23.974852 * X1 + 29.007851
    A133 = 0.237096 * X1 ** 3 - 3.535062 * X1 ** 2 + 16.368376 * X1 - 20.539908
    A134 = -0.067119 * X1 ** 3 + 0.966362 * X1 ** 2 - 4.407535 * X1 + 5.894703
    A11 = A111 * X2 ** 2 + A112 * X2 + A113
    A12 = A121 * X2 ** 3 + A122 * X2 ** 2 + A123 * X2 + A124
    A13 = A131 * X2 ** 3 + A132 * X2 ** 2 + A133 * X2 + A134
    AA111 = 17.945 * X1 ** 3 - 166.294 * X1 ** 2 + 489.799 * X1 - 493.142
    AA112 = -25.507 * X1 ** 3 + 236.275 * X1 ** 2 - 698.683 * X1 + 701.494
    AA113 = 9.077 * X1 ** 3 - 84.332 * X1 ** 2 + 249.983 * X1 - 250.787
    AA121 = -16.872 * X1 ** 3 + 156.399 * X1 ** 2 - 460.689 * X1 + 463.848
    AA122 = 24.015 * X1 ** 3 - 222.507 * X1 ** 2 + 658.027 * X1 - 660.665
    AA123 = -8.56 * X1 ** 3 + 79.549 * X1 ** 2 - 235.827 * X1 + 236.579
    AA11 = AA111 * X2 ** 2 + AA112 * X2 + AA113
    AA12 = AA121 * X2 ** 2 + AA122 * X2 + AA123
    AA1 = (AA11 * X3 + AA12) * (1 - X4) + 1.0
    A1 = (A11 * X4 ** 2 + A12 * X4 + A13) * AA1
    A2 = -1.402 * X4 ** 3 + 7.189 * X4 ** 2 - 10.993 * X4 + 9.45
    A31 = -7686.0287 * X2 ** 6 + 30131.5678 * X2 ** 5 - 49048.9664 * X2 ** 4 + 42480.7709 * X2 ** 3 - 20665.147 * X2 ** 2 + 5355.2035 * X2 - 577.8827
    A32 = 61639.9103 * X2 ** 6 - 241201.0598 * X2 ** 5 + 392579.5937 * X2 ** 4 - 340629.4699 * X2 ** 3 + 166348.6917 * X2 ** 2 - 43358.7938 * X2 + 4714.7918
    A33 = -130677.4903 * X2 ** 6 + 507996.2604 * X2 ** 5 - 826728.7127 * X2 ** 4 + 722677.104 * X2 ** 3 - 358360.7392 * X2 ** 2 + 95501.4948 * X2 - 10682.8619
    A34 = -110034.6584 * X2 ** 6 + 446051.22 * X2 ** 5 - 724186.4643 * X2 ** 4 + 599411.9264 * X2 ** 3 - 264294.7189 * X2 ** 2 + 58039.7328 * X2 - 4774.6414
    A35 = 709672.0656 * X2 ** 6 - 2803850.2395 * X2 ** 5 + 4553780.5017 * X2 ** 4 - 3888378.9905 * X2 ** 3 + 1839829.259 * X2 ** 2 - 457313.6939 * X2 + 46600.823
    A36 = -822735.9289 * X2 ** 6 + 3238899.7308 * X2 ** 5 - 5256636.5472 * X2 ** 4 + 4500543.147 * X2 ** 3 - 2143487.3508 * X2 ** 2 + 538548.1194 * X2 - 55751.1528
    A37 = 299122.8727 * X2 ** 6 - 1175773.1606 * X2 ** 5 + 1907356.1357 * X2 ** 4 - 1634256.8172 * X2 ** 3 + 780020.9393 * X2 ** 2 - 196679.7143 * X2 + 20467.0904
    AA311 = (
                        -17.102 * X2 ** 3 + 41.495 * X2 ** 2 - 33.234 * X2 + 8.8007) * X4 + 36.566 * X2 ** 3 - 89.203 * X2 ** 2 + 71.8 * X2 - 18.108
    AA31 = (-0.3767 * X1 ** 3 + 3.39 * X1 ** 2 - 10.356 * X1 + 11.588) * AA311
    AA32 = -0.0727 * X1 ** 2 + 0.7 * X1 - 1.2818
    XX4 = X4 - AA32
    AA3 = AA31 * (
                -1.05584 * XX4 ** 9 + 12.688 * XX4 ** 8 - 63.70534 * XX4 ** 7 + 172.84571 * XX4 ** 6 - 274.05701 * XX4 ** 5 + 257.68705 * XX4 ** 4 - 141.40915 * XX4 ** 3 + 44.13177 * XX4 ** 2 - 7.1654 * XX4 - 0.0495 * X1 ** 2 + 0.4518 * X1 - 0.61655)
    A3 = A31 * X4 ** 6 + A32 * X4 ** 5 + A33 * X4 ** 4 + A34 * X4 ** 3 + A35 * X4 ** 2 + A36 * X4 + A37 + AA3
    BWHAT = A1 / X5 * EXP(-A2 * (LOG(X5) - A3) ** 2 / 1.44)
    return BWHAT


def calculate_B_F(BD, BRTH, CB, DRAFT, KVC, LPP, OGD, OMEGA, PHI):
    """
    Skin friction damping B_F

    :param BD: Beam/DRAFT
    :param BRTH: Beam [m]
    :param CB: Block coefficient [-]
    :param DRAFT: DRAFT : ship draught [m]
    :param KVC = 1.14e-6  # Kinematic Viscosity Coefficient
    :param LPP: [m]
    :param OGD: OG/DRAFT
    :param OMEGA: Frequency of motion [rad/s]
    :param PHI: Roll angle [deg]
    :param CMID: Middship coefficient (A_m/(B*d) [-]

    :return: B_F
    """

    # There must be a typo in the original fortran code it was 102 instead of 1025!?
    RO = 1025  # Density of water

    TW = 2 * PI / OMEGA
    RF = DRAFT * ((0.887 + 0.145 * CB) * (1.7 + CB * BD) - 2.0 * OGD) / PI
    SF = LPP * (1.75 * DRAFT + CB * BRTH)
    CF = 1.328 * ((3.22 * RF ** 2 * (PHI * PI / 180) ** 2) / (TW * KVC)) ** -0.5
    BF = 4.0 / 3.0 / PI * RO * SF * RF ** 3 * (PHI * PI / 180) * OMEGA * CF
    BFHAT = BF / (RO * LPP * BRTH ** 3 * DRAFT * CB) * SQRT(BRTH / 2.0 / 9.81)
    return BFHAT


class SimplifiedIkedaInputError(ValueError): pass

limits_kawahara = {
    'CB'    : (0.5,0.85),
    r'B/d'  : (2.5,4.5),
    r'OG/d' : (-1.5,0.2),
    'CMID'    : (0.9,0.99),
    r'bBk/B': (0.01, 0.06),
    r'lBk/LPP': (0.05, 0.4),
    'OMEGA_hat': (0,1.0)
}  # Input limits for damping according to the original paper ny Kawahara

def _calculate_limit_compare_value(LPP, Beam, OG, lBK, bBK, OMEGA,
                                   DRAFT):
    inputs = {
        r'B/d': Beam / DRAFT,
        r'OG/d': OG / DRAFT,
        r'bBk/B': bBK / Beam,
        r'lBk/LPP': lBK / LPP,
        'OMEGA_hat': OMEGA * SQRT(Beam / 2 / 9.81)
    }
    return inputs

def verify_inputs(LPP,Beam,CB,CMID,OG,PHI,lBK,bBK,OMEGA,
                           DRAFT):
    inputs = {
        'LPP': LPP,
        'Beam': Beam,
        'CB': CB,
        'CMID': CMID,
        'OG': OG,
        'PHI': PHI,
        'BKL': lBK,
        'BKB': bBK,
        'OMEGA': OMEGA,
        'DRAFT': DRAFT,
    }
    limit_inputs = _calculate_limit_compare_value(LPP, Beam, OG, lBK, bBK, OMEGA,
                                                  DRAFT)

    inputs.update(limit_inputs)

    exclude_zero = 10*-10
    limits={
        'LPP'   : (exclude_zero,np.inf),
        'Beam'  : (exclude_zero,np.inf),
        'CB'    : (0.4,1.0),
        'CMID'  : (0.4,1.0),
        'OG'    : (-np.inf,np.inf),
        'PHI'   : (exclude_zero, np.inf),
        'BKL'   : (0,np.inf),
        'BKB'   : (0,np.inf),
        'OMEGA' : (exclude_zero,np.inf),
        'DRAFT' : (exclude_zero,np.inf)
    }
    limits.update(limits_kawahara)
    if lBK==0:
        # Remove bilge keel limit when the bilge keel is not present:
        if r'bBk/B' in limits:
            limits.pop(r'bBk/B')

        if r'lBk/LPP' in limits:
            limits.pop(r'lBk/LPP')

    for key,value in inputs.items():

        if not key in limits:
            continue

        lims = limits.get(key)

        if np.any(pd.isnull(value)):
            raise SimplifiedIkedaInputError('%s is NaN' % key)

        #if not np.all((lims[0] <= value) & (value <= lims[1])):
        #    raise SimplifiedIkedaInputError('%s has a bad value:%f is not in (%f,%f)' % (key,value, lims[0], lims[1]))

        if np.any((value - lims[0]) < -0.000001):
            raise SimplifiedIkedaInputError('%s:%f is too small  (<%f)' % (key, value, lims[0]))

        if np.any((lims[1] - value) < -0.000001):
            raise SimplifiedIkedaInputError('%s:%f is too large  (>%f)' % (key, value, lims[1]))

    if np.any((LPP/Beam > 100)):
        raise SimplifiedIkedaInputError('Lpp/Beam has bad ratio' % (LPP/Beam))

def _calculate_limit_value(LPP, Beam, DRAFT):

    beam_limit = np.array(limits_kawahara['B/d']) * DRAFT
    _beam=Beam
    if Beam < beam_limit[0]:
        _beam =beam_limit[0]
    elif Beam > beam_limit[1]:
        _beam = beam_limit[1]

    limits = {
        'CB': np.array(limits_kawahara['CB']),
        'CMID': np.array(limits_kawahara['CMID']),
        'Beam': np.array(limits_kawahara['B/d'])*DRAFT,
        'OG': np.array(limits_kawahara[r'OG/d'])*DRAFT,
        'BKB': np.array(limits_kawahara[r'bBk/B'])*_beam,
        'BKL': np.array(limits_kawahara[r'lBk/LPP'])*LPP,
        'OMEGA': np.array(limits_kawahara['OMEGA_hat'])/(SQRT(_beam / 2 / 9.81)),
    }

    return limits

def _limit_inputs(LPP,Beam,CB,CMID,OG,PHI,lBK,bBK,OMEGA,
                           DRAFT):
    inputs = {
        'LPP': LPP,
        'Beam': Beam,
        'CB': CB,
        'CMID': CMID,
        'OG': OG,
        'PHI': PHI,
        'BKL': lBK,
        'BKB': bBK,
        'OMEGA': OMEGA,
        'DRAFT': DRAFT,
    }

    limits = _calculate_limit_value(LPP=LPP, Beam=Beam, DRAFT=DRAFT)

    if lBK==0:
        # Remove bilge keel limit when the bilge keel is not present:
        if 'BKL' in limits:
            limits.pop('BKL')

        if 'BKB' in limits:
            limits.pop('BKB')

    outputs = {}
    for key, value in inputs.items():

        if not key in limits:
            outputs[key] = inputs[key]
            continue

        limit = limits[key]
        if inputs[key] < limit[0]:
            outputs[key] = limit[0]
        elif inputs[key] > limit[1]:
            outputs[key] = limit[1]
        else:
            outputs[key] = inputs[key]

    return outputs




from numpy import sqrt,pi,tanh, exp, cos, sin, arccos
import numpy as np

def eddy_sections(bwl:np.ndarray, a_1:np.ndarray, a_3:np.ndarray, sigma:np.ndarray, H0:np.ndarray, Ts:np.ndarray,
         OG:float, R:float, wE:float, fi_a:float, ra=1000.0):
    """
    Calculation of eddy damping according to
    Ikeda, Y.,
    1978. On eddy making component of roll damping force on naked hull. University of Osaka Prefacture,
    Department of Naval Architecture, Japan, Report No. 00403,
    Published in: Journal of Society of Naval Architects of Japan, Volume 142.

    Parameters
    ----------
    bwl
        sectional beam water line [m]
    a_1
        sectional lewis coefficients
    a_3
        sectional lewis coefficients
    sigma
        sectional coefficient
    H0
        sectional coefficient
    Ts
        sectional draft [m]
    OG
        vertical distance water line to cg [m]
    R
        bilge radius [m]
    ra
        water density [kg/m3]
    wE
        roll requency [rad/s]
    fi_a
        roll amplitude [rad]

    Returns
    -------
    B_e0s
        Eddy damping per unit length for the sections at zero speed.
    """

    d = Ts

    C_r = calculate_C_r(bwl=bwl,a_1=a_1, a_3=a_3, sigma=sigma, H0=H0, d=Ts, OG=OG, R=R, ra=ra)

    B_e0s = 4*ra/(3*pi)*d**4*wE*fi_a*C_r  # (6) (Rewritten)

    return np.array([B_e0s])

def calculate_C_r(bwl:np.ndarray, a_1:np.ndarray, a_3:np.ndarray, sigma:np.ndarray, H0:np.ndarray, d,
         OG:float, R:float, ra=1000.0):

    """
        Parameters
    ----------
    bwl
        sectional beam water line [m]
    a_1
        sectional lewis coefficients
    a_3
        sectional lewis coefficients
    sigma
        sectional coefficient
    H0
        sectional coefficient
    Ts
        sectional draft [m]
    OG
        vertical distance water line to cg [m]
    R
        bilge radius [m]
    ra
        water density [kg/m3]

    Returns
    -------
    B_e0s
        Eddy damping per unit length for the sections at zero speed.
    """


    ## start to obtain y from the parameter H0 and a of the cylinder section:
    gamma, r_max = calculate_gamma(sigma=sigma, OG=OG, d=d, a_1=a_1, a_3=a_3, H0=H0, bwl=bwl)

    C_p = calculate_C_p(gamma=gamma)  # (14)

    f_1 = 1/2*(1 + tanh(20*(sigma - 0.7)))  # (12)
    f_2 = calculate_f2(sigma=sigma)  # (15)

    P_m_ = 3*1/2*ra*r_max**2*C_p  # (13) (*|phi1d|*phi2d left out, but will cancel out in the next)
    C_r = 2/(ra*d**2)*((1 - f_1*R/d)*(1 - OG/d - f_1*R/d) + f_2*(H0 - f_1*R/d)**2)*P_m_/3  # (10)
    
    #M_re = 1/2*ra*r_max**2*d**2*C_p*((1-f_1*R/d)*(1 - OG/d - f_1*R/d) + f_2*(H0 - f_1*R/d)**2)
    #C_r = M_re/(1/2*ra*d**4)
    
    return C_r


def calculate_f2(sigma):
    f_2 = 1 / 2 * (1 - cos(pi * sigma)) - 1.5 * (1 - exp(-5 * (1 - sigma))) * sin(pi * sigma) ** 2  # (15)
    return f_2

def calculate_C_p(gamma):
    C_p = 1 / 2 * (0.87 * exp(-gamma) - 4 * exp(-0.187 * gamma) + 3)  # (14)
    return C_p

def calculate_gamma(sigma, OG, d, a_1, a_3, H0, bwl):

    # (A-1):
    sigma_ = (sigma-OG/d)/(1-OG/d)
    H0_ = H0/(1-OG/d)

    # (A-4)
    def calculate_A(psi):
        return -2 * a_3 * cos(5 * psi) + a_1 * (1 - a_3) * cos(3 * psi) + (
            (6 - 3 * a_1) * a_3 ** 2 + (a_1 ** 2 - 3 * a_1) * a_3 + a_1 ** 2) * cos(psi)

    def calculate_B(psi):
        return -2 * a_3 * sin(5 * psi) + a_1 * (1 - a_3) * sin(3 * psi) + (
            (6 + 3 * a_1) * a_3 ** 2 + (3 * a_1 + a_1 ** 2) * a_3 + a_1 ** 2) * sin(psi)

    def calculate_M():
        return bwl / (2 * (1 + a_1 + a_3))  # Note B is bwl!!!

    def calculate_H(psi):
        return 1 + a_1 ** 2 + 9 * a_3 ** 2 + 2 * a_1 * (1 - 3 * a_3) * cos(2 * psi) - 6 * a_3 * cos(4 * psi)

    def calculate_r_max(psi):

        M = calculate_M()

        return M * sqrt(
            ((1 + a_1) * sin(psi) - a_3 * sin(3 * psi)) ** 2 + (
                        (1 - a_1) * cos(psi) + a_3 * cos(3 * psi)) ** 2)  # (A-5)

    # (A-6)
    psi_1 = 0
    factor = a_1*(1+a_3)/(4*a_3)  
    # Limit factor to [-1,1]
    mask = np.abs(factor) > 1
    factor[mask] = 1*np.sign(factor[mask])
    psi_2 = 1/2*arccos(factor)

    # (A-7)
    r_max_1 = calculate_r_max(psi_1)
    r_max_2 = calculate_r_max(psi_2)
    mask = r_max_1 >= r_max_2
    psi = np.array(psi_2)
    psi[mask] = psi_1

    A = calculate_A(psi=psi)
    B = calculate_B(psi=psi)
    M = calculate_M()
    H = calculate_H(psi=psi)

    r_max = np.array(r_max_2)
    r_max[mask] = r_max_1[mask]

    f_3 = 1 + 4*exp(-1.65*10**5*(1-sigma)**2)  # (A-9)

    gamma = sqrt(pi)*f_3/((2*d)*(1-OG/d)*sqrt(H0_*sigma_))*(r_max+2*M/H)*sqrt(A**2+B**2)  # (A-10)

    return gamma, r_max



import numpy as np
from sklearn.base import clone
from sklearn.model_selection import KFold
import matplotlib.pyplot as plt

def model_filter(group, models):
    return group.name in models

def cross_validates(model, data, features, label='B_e_hat', n_splits=5, itterations=10):
    scores = []
    for i in range(itterations):
        scores_ = cross_validate(model, data, features=features, label=label, n_splits=n_splits)
        scores.append(scores_)

    return np.array(scores)


def cross_validate(model, data, features, label='B_e_hat', n_splits=5):
    groups_model = data.groupby(by='model_number')
    models = data['model_number'].unique()
    np.random.shuffle(models)  # Inplace

    kf = KFold(n_splits=n_splits)
    scores = []
    model_test = clone(model)
    for train_index, test_index in kf.split(models):
        models_train = models[train_index]
        models_test = models[test_index]

        data_train = groups_model.filter(func=model_filter, models=models_train)
        data_test = X_test = groups_model.filter(func=model_filter, models=models_test)

        X_train = data_train[features]
        X_test = data_test[features]

        y_train = data_train[label]
        y_test = data_test[label]

        model_test.fit(X=X_train, y=y_train)
        score = model_test.score(X=X_test, y=y_test)
        scores.append(score)


    return scores


def plot_validate(model, data, features, label='B_e_hat', n_splits=5):
    groups_model = data.groupby(by='model_number')
    models = data['model_number'].unique()
    np.random.shuffle(models)  # Inplace

    kf = KFold(n_splits=n_splits)
    model_test = clone(model)

    nrows=2
    ncols = int(np.ceil(n_splits/nrows))

    fig,axes=plt.subplots(ncols=ncols, nrows=nrows)

    axess=axes.flatten()

    not_used = axess[n_splits:]
    for ax in not_used:
        ax.remove()

    axess=axess[0:n_splits]


    fold=0
    for (train_index, test_index),ax in zip(kf.split(models),axess):
        models_train = models[train_index]
        models_test = models[test_index]

        data_train = groups_model.filter(func=model_filter, models=models_train)
        data_test = X_test = groups_model.filter(func=model_filter, models=models_test)

        X_train = data_train[features]
        X_test = data_test[features]

        y_train = data_train[label]
        y_test = data_test[label]

        model_test.fit(X=X_train, y=y_train)
        ax.plot(y_test, model_test.predict(X=X_test),'.')

        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        lim = np.max([xlim[1], ylim[1]])
        ax.set_xlim(0, lim)
        ax.set_ylim(0, lim)
        ax.set_title('Test fold %i' % fold)
        fold+=1
        ax.plot([0, lim], [0, lim], 'r-')
        ax.set_aspect('equal', 'box')
        ax.grid(True)

    #axes[0].legend()
    axess[0].set_ylabel('$\hat{B_e}$ (prediction)')
    axess[3].set_ylabel('$\hat{B_e}$ (model test)')
    axess[3].set_xlabel('$\hat{B_e}$ (model test)')
    axess[4].set_xlabel('$\hat{B_e}$ (model test)')

    return fig


#! /usr/bin/env python
"""A template for scikit-learn compatible packages."""

import codecs
import os

from setuptools import find_packages, setup

# get __version__ from _version.py
ver_file = os.path.join('rolldecayestimators', '_version.py')
with open(ver_file) as f:
    exec(f.read())

DISTNAME = 'rolldecay-estimators'
DESCRIPTION = 'A template for scikit-learn compatible packages.'
with open("README.md", "r") as fh:
    long_description = fh.read()

MAINTAINER = 'Martin Alexandersson'
MAINTAINER_EMAIL = 'maralex@chalmers.se'
URL = 'https://github.com/scikit-learn-contrib/project-template'
LICENSE = 'new BSD'
DOWNLOAD_URL = 'https://github.com/martinlarsalbert/rolldecay-estimators'
VERSION = __version__
INSTALL_REQUIRES = ['numpy', 'scipy', 'scikit-learn','pandas','sympy','matplotlib','dill']
CLASSIFIERS = ['Intended Audience :: Science/Research',
               'Intended Audience :: Developers',
               'License :: OSI Approved',
               'Programming Language :: Python',
               'Topic :: Software Development',
               'Topic :: Scientific/Engineering',
               'Operating System :: Microsoft :: Windows',
               'Operating System :: POSIX',
               'Operating System :: Unix',
               'Operating System :: MacOS',
               'Programming Language :: Python :: 2.7',
               'Programming Language :: Python :: 3.5',
               'Programming Language :: Python :: 3.6',
               'Programming Language :: Python :: 3.7']
EXTRAS_REQUIRE = {
    'tests': [
        'pytest',
        'pytest-cov'],
    'docs': [
        'sphinx',
        'sphinx-gallery',
        'sphinx_rtd_theme',
        'numpydoc',
        'matplotlib'
    ]
}

package_data= {
          "rolldecayestimators": ["*.csv"],
      }

setup(name=DISTNAME,
      maintainer=MAINTAINER,
      maintainer_email=MAINTAINER_EMAIL,
      description=DESCRIPTION,
      license=LICENSE,
      url=URL,
      version=VERSION,
      download_url=DOWNLOAD_URL,
      long_description=long_description,
      zip_safe=False,  # the package can run out of an .egg file
      classifiers=CLASSIFIERS,
      packages=find_packages(),
      install_requires=INSTALL_REQUIRES,
      extras_require=EXTRAS_REQUIRE,
      package_data=package_data,
      author="Martin Alexandersson",
      author_email='maralex@chalmers.se',
      python_requires='>=3.5',
      keywords='rolldecayestimators',
      )



import pytest
import pandas as pd
import numpy as np
from numpy import pi, sqrt
import matplotlib.pyplot as plt
import os
from numpy.testing import assert_almost_equal, assert_allclose

from rolldecayestimators.ikeda import Ikeda, IkedaR
from rolldecayestimators import lambdas
import rolldecayestimators
import pyscores2.test
import pyscores2.indata
import pyscores2.output

@pytest.fixture
def ikeda():

    N=10
    data = np.zeros(N)
    w_hat=np.linspace(0.1,1, N)
    B_W0_hat = pd.Series(data=data, index=w_hat)
    fi_a = np.deg2rad(10)
    beam=10
    lpp=100
    kg = 1
    Cb=0.7
    draught = 10
    volume = Cb*lpp*draught*beam
    V = 5
    w = 0.2
    A0 = 0.95

    N_sections = 21
    x_s = np.linspace(0, lpp, N_sections)
    data = {
        'B_s' : beam*np.ones(N_sections),
        'T_s' : draught*np.ones(N_sections),
        'C_s' : A0*np.ones(N_sections),
    }
    sections=pd.DataFrame(data=data, index=x_s)

    i= Ikeda(V=V, w=w, B_W0_hat=B_W0_hat, fi_a=fi_a, beam=beam, lpp=lpp, kg=kg, volume=volume,
             sections=sections, BKL=0, BKB=0)
    i.R=2.0  # Set bilge radius manually

    yield i

@pytest.fixture
def ikeda_faust():
    # this is indata from Carl-Johans matlab example for ship: Faust.

    ScaleF = 1  # %/29.565;                # Scale Factor [-]
    visc = 1.15 * 10 ** -6;  # [m2/s], kinematic viscosity
    Cb = 0.61;  # Block coeff
    L = 220 * ScaleF;  # Length
    vcg = 14.4 * ScaleF;  # roll axis (vertical centre of gravity) [m]
    vcg = 14.9 * ScaleF;  # roll axis (vertical centre of gravity) [m]
    B = 32.26 * ScaleF;  # Breadth of hull [m]
    d = 9.5 * ScaleF;  # Draught of hull [m]
    A = 0.93 * B * d;  # Area of cross section of hull [m2]
    bBK = 0.4 * ScaleF;  # breadth of Bilge keel [m] !!(=height???)
    R = 5 * ScaleF;  # Bilge Radis
    g = 9.81;
    C_mid = 0.93;

    OG = -1 * (vcg - d)  # *0.8;                    # distance from roll axis to still water level
    Ho = B / (2 * d);  # half breadth to draft ratio
    ra = 1025;  # density of water

    # locals
    LBK = L / 4;  # Approx
    disp = L * B * d * Cb;  # Displacement

    # variables!!
    T = 27.6 * sqrt(ScaleF);
    wE = 2 * pi * 1 / T;  # circular frequency
    fi_a = 10 * pi / 180;  # roll amplitude !!rad??
    V = 0;  # Speed

    data_path_faust = os.path.join(rolldecayestimators.path, 'Bw0_faust.csv')
    data_faust = pd.read_csv(data_path_faust, sep=';')
    data_faust['w_hat'] = lambdas.omega_hat(beam=B, g=g, omega0=data_faust['w_vec'])
    data_faust['B_W0_hat'] = lambdas.B_hat_lambda(B=data_faust['b44_vec'], Disp=disp, beam=B, g=g, rho=ra)

    data_faust.set_index('w_hat', inplace=True)
    B_W0_hat = data_faust['B_W0_hat']


    N_sections = 21
    x_s = np.linspace(0, L, N_sections)
    data = {
        'B_s': B * np.ones(N_sections),
        'T_s': d * np.ones(N_sections),
        'C_s': C_mid*np.ones(N_sections),
    }
    sections = pd.DataFrame(data=data, index=x_s)  # Fake sections (not testing the eddy)

    i = Ikeda(V=V, w=wE, B_W0_hat=B_W0_hat, fi_a=fi_a, beam=B, lpp=L, kg=vcg, volume=disp,
              sections=sections, BKB=bBK, BKL=LBK)
    i.R = R  # Set bilge radius manually

    yield i

def test_R(ikeda):
    assert ikeda.R==2.0

def test_calculate_Ikeda(ikeda):
    B_44=ikeda.calculate_B44()

def test_calculate_Ikeda_faust(ikeda_faust):
    B_44=ikeda_faust.calculate_B44()

def test_Bw0(ikeda_faust):
    Bw0=ikeda_faust.calculate_B_W0()
    assert_allclose(Bw0, 5.541101e-05, rtol=0.001)


def test_bw44_V0(ikeda_faust):

    ikeda_faust.V = 0  ## Ship speed
    bw44 = ikeda_faust.calculate_B_W()
    assert_allclose(bw44, 5.541101e-05, rtol=0.01)


def test_bilge_keel(ikeda_faust):

    ikeda_faust.V = 0  ## Ship speed
    T = 27.6
    ikeda_faust.w = 2 * pi * 1 / T;  # circular frequency
    ikeda_faust.fi_a = 10 * pi / 180;  # roll amplitude !!rad??

    B_BK = ikeda_faust.calculate_B_BK()

    assert_allclose(B_BK, ikeda_faust.B_hat(75841485), rtol=0.01)

@pytest.mark.skip('This one does not work yet')
def test_friction(ikeda_faust):
    ikeda_faust.V = 0  ## Ship speed
    T = 27.6
    ikeda_faust.w = 2 * pi * 1 / T;  # circular frequency
    ikeda_faust.fi_a = 10 * pi / 180;  # roll amplitude !!rad??

    B44F = ikeda_faust.calculate_B_F()

    assert_allclose(B44F, ikeda_faust.B_hat(5652721), rtol=0.001)


def test_hull_lift(ikeda_faust):

    ikeda_faust.V = 10  ## Ship speed
    B44L = ikeda_faust.calculate_B_L()
    assert_allclose(B44L, ikeda_faust.B_hat(1.692159e+08), rtol=0.001)

@pytest.fixture
def indata():
    indata=pyscores2.indata.Indata()
    indata.open(indataPath=pyscores2.test.indata_path)
    yield indata

@pytest.fixture
def output():
    output=pyscores2.output.OutputFile(filePath=pyscores2.test.outdata_path)
    yield output

def test_load_scoresII(indata, output):

    V = 5
    w = 0.2
    fi_a = np.deg2rad(10)

    ikeda = Ikeda.load_scoresII(indata=indata, output_file=output, V=V, w=w, fi_a=fi_a, BKB=0, BKL=0, kg=0)
    ikeda.R = 2.0  # Set bilge radius manually
    B_44_hat = ikeda.calculate_B44()

def test_calculate_R_b(indata, output):

    V = 5
    w = 0.2
    fi_a = np.deg2rad(10)

    ikeda = Ikeda.load_scoresII(indata=indata, output_file=output, V=V, w=w, fi_a=fi_a, BKB=0, BKL=0, kg=0)
    R_b = ikeda.calculate_R_b()


def test_load_scoresII_scale(indata, output):

    V = 5
    w = 0.2
    fi_a = np.deg2rad(10)
    R=2.0

    ikeda = Ikeda.load_scoresII(indata=indata, output_file=output, V=V, w=w, fi_a=fi_a, BKB=0, BKL=0, kg=0)
    ikeda.R = R  # Set bilge radius manually

    scale_factor=50
    V_m=V/np.sqrt(scale_factor)
    w_m=w*np.sqrt(scale_factor)

    ikeda_model = Ikeda.load_scoresII(indata=indata, output_file=output, V=V_m, w=w_m, fi_a=fi_a, BKB=0, BKL=0,
                                      scale_factor=scale_factor, kg=0)
    ikeda_model.R = R/scale_factor  # Set bilge radius manually

    assert_almost_equal(ikeda.calculate_B_W(),  ikeda_model.calculate_B_W() )
    assert_almost_equal(ikeda.calculate_B_BK(), ikeda_model.calculate_B_BK())
    assert_almost_equal(ikeda.calculate_B_E(),  ikeda_model.calculate_B_E() )
    assert_almost_equal(ikeda.calculate_B_L(),  ikeda_model.calculate_B_L() )

def test_load_scoresII_scale_V_variation(indata, output):

    scale_factor = 68
    N = 200
    V = np.linspace(0, 15.5, N) * 1.852 / 3.6 / np.sqrt(scale_factor)
    kg = 0.2735294117647059
    w = 2.4755750032144674

    ## Load ScoresII results
    phi_a_deg = 10
    phi_a = np.deg2rad(phi_a_deg) * np.ones(N)

    ikeda_estimator = Ikeda.load_scoresII(V=V,
                                          w=w,
                                          fi_a=phi_a,
                                          indata=indata,
                                          output_file=output,
                                          scale_factor=scale_factor, BKL=0, BKB=0, kg=kg)

    results = ikeda_estimator.calculate()
    results['V'] = V
    results.set_index('V', inplace=True)

    phi_as = np.deg2rad(np.linspace(0, phi_a_deg, N))
    ikeda_estimator2 = Ikeda.load_scoresII(V=np.max(V),
                                           w=w,
                                           fi_a=phi_as,
                                           indata=indata,
                                           output_file=output,
                                           scale_factor=scale_factor, BKL=0, BKB=0, kg=kg)

    results2 = ikeda_estimator2.calculate()
    results2['phi_a'] = phi_as
    results2.set_index('phi_a', inplace=True)

    fig,axes=plt.subplots(ncols=2)

    ax=axes[0]
    results.plot(y='B_44_hat',ax=ax)

    ax = axes[1]
    results2.plot(y='B_44_hat', ax=ax)

    fig.show()

    assert_almost_equal(results.iloc[-1]['B_44_hat'], results2.iloc[-1]['B_44_hat'])



"""
This module contain a class to calculate Ikeda in various ways
"""
import numpy as np
import pandas as pd
from numpy import sqrt as sqrt
from scipy.integrate import simps

from rolldecayestimators import ikeda_speed
from rolldecayestimators import ikeda_barge
from rolldecayestimators import simplified_ikeda
from rolldecayestimators import lambdas

class SectionsError(ValueError): pass
class BilgeKeelError(ValueError): pass
class InputError(ValueError): pass
class WInputError(InputError): pass
class PhiAInputError(InputError): pass


import pyscores2.indata
import pyscores2.output

class Ikeda():
    """
    Class that helps with running various calculations known as the "Ikedas method" to predict ship roll damping.
    The idea with this class is that it can be subclassed so that varius roll damping contributions can be replaced or
    removed in order to study different "flavors" of the Ikeda methods (I have seen a lot of different implementations).


    """
    def __init__(self, V:np.ndarray, w:np.ndarray, fi_a:float, B_W0_hat:pd.Series, beam:float, lpp:float,
                 kg:float, volume:float, sections:pd.DataFrame, BKL, BKB, R_b=None,
                 g=9.81, rho=1000.0, visc =1.15*10**-6, scale_factor=1.0, S_f=None, sigma_limit = 0.99, **kwargs):
        """
        Manually specify the inputs to the calculations.
        Note: Some inputs need to be specified here, but others can be defined in other ways
        (by importing a ship geometry etc.)

        Parameters
        ----------
        V
            ship speed [m/s]
        w
            roll frequency [rad/s]
        fi_a
            roll amplitude [rad]
        B_W0_hat : pd.Series
            values : wave roll damping for various frequencies [-]
            index: frequencies hat [-]
        beam
            ship beam [m]
        lpp
            ship perpedicular length [m]
        kg
            vertical centre of gravity [m] (positive upward)
        volume
            ship displaced volume [m3]
        sections
            ship hull geometry sections defined in a pandas data frame:
            sections.index  : section x coordinate measured from AP [m]
            sections['B_s'] : sectional beam [m] (incl. both sides)
            sections['T_s'] : sectional draught [m]
            sections['C_s'] :   sectional area coefficient [-]
                                C_s = S_s/(B_s*T_s)
        BKL
            length bilge keel [m] (=0 --> no bilge keel)
        BKB
            height bilge keel [m] (=0 --> no bilge keel)
        scale_factor : float
            scale factor is used to calculate the wetted surface in the skin friction damping B_F
        g
            gravity [m/s2]
        rho
            water density [kg/m3]
        S_f
            Wetted surface (optional) [m2]
        sigma_limit
            Upper limit to the section area coefficient default: 0.99 (In accordance to Kawahara)

            Kawahara, Y., Maekawa, K., Ikeda, Y.,
            2011. A Simple Prediction Formula of Roll Damping of Conventional Cargo Ships on the Basis of Ikeda’s Method and Its Limitation,
            in: Almeida Santos Neves, M., Belenky, V.L., de Kat, J.O., Spyrou, K., Umeda, N. (Eds.),
            Contemporary Ideas on Ship Stability and Capsizing in Waves, Fluid Mechanics and Its Applications. Springer Netherlands,
            Dordrecht, pp. 465–486. https://doi.org/10.1007/978-94-007-1482-3_26

        R_b float
            Bilge radius for stations [m] (optional)

        Returns
        -------
        None
        """
        self.V = np.array(V, dtype=float)
        self.g = g
        self.w =  w
        self.fi_a =  fi_a
        self.B_W0_hat= B_W0_hat
        self.beam= beam
        self.lpp= lpp
        self.kg= kg
        self.volume =  volume
        self.sections= sections
        self.lBK= BKL
        self.bBK= BKB
        self.rho= rho
        self.visc= visc
        self.scale_factor= scale_factor
        self.S_f = S_f
        self.sigma_limit = sigma_limit

        if not (R_b is None):
            self._R = R_b

    @classmethod
    def load_scoresII(cls, V:np.ndarray, w:np.ndarray, fi_a:float, indata:pyscores2.indata.Indata,
                      output_file:pyscores2.output.OutputFile, BKL: float, BKB :float, kg, g=9.81, rho=1000.0, visc =1.15 * 10 ** -6,
                      scale_factor=1.0, S_f = None, sigma_limit = 0.99, R_b = None,
                      **kwargs):
        """
        Creaate a object from indata and output from ScoresII

        Parameters
        ----------
        V
        w
        fi_a
        indata : pyscores2.indata.Indata
            Indata to ScoresII program
        output_file :pyscores2.output.OutputFile
            Outdata from ScoresII program
        BKL
            bilge keel length [m]
        BKB
            bilge keel height [m]
        kg
            vertical centre of gravity [m] (positive upward)
        g
        rho
        visc
        kwargs
        scale_factor : float
            scale factor is used to calculate the wetted surface in the skin friction damping B_F
        S_f
            Wetted surface (optional) [m2]
        sigma_limit
            Upper limit to the section area coefficient default: 0.99 (In accordance to Kawahara)

            Kawahara, Y., Maekawa, K., Ikeda, Y.,
            2011. A Simple Prediction Formula of Roll Damping of Conventional Cargo Ships on the Basis of Ikeda’s Method and Its Limitation,
            in: Almeida Santos Neves, M., Belenky, V.L., de Kat, J.O., Spyrou, K., Umeda, N. (Eds.),
            Contemporary Ideas on Ship Stability and Capsizing in Waves, Fluid Mechanics and Its Applications. Springer Netherlands,
            Dordrecht, pp. 465–486. https://doi.org/10.1007/978-94-007-1482-3_26

        R_b float
            Bilge radius for stations [m] (optional)

        Returns
        -------

        """

        N_sections = len(indata.bs)
        lpp = indata.lpp/scale_factor
        x_s = np.linspace(0, lpp, N_sections)
        data = {
            'B_s': indata.bs,
            'T_s': indata.ts,
            'C_s': indata.cScores,
        }
        sections = pd.DataFrame(data=data, index=x_s)
        sections['B_s']/=scale_factor
        sections['T_s'] /= scale_factor
        beam=sections['B_s'].max()
        volume=indata.displacement/(scale_factor**3)
        draught=sections['T_s'].max()
        zcg=indata.zcg/scale_factor
        #kg=(zcg+draught)

        ws, data = output_file.calculate_B_W0()  # Full scale values
        beam_fullscale=beam*scale_factor
        volume_fullscale=volume*scale_factor**3
        ws_hat = lambdas.omega_hat(beam=beam_fullscale, g=g, omega0=ws)
        B_W0_hat_ = lambdas.B_hat_lambda(B=data, Disp=volume_fullscale, beam=beam_fullscale, g=g, rho=rho)
        B_W0_hat = pd.Series(data=B_W0_hat_, index=ws_hat)

        return cls(V=V, w=w, fi_a=fi_a, B_W0_hat=B_W0_hat, beam=beam, lpp=lpp, kg=kg, volume=volume,
                   sections=sections, BKL=BKL, BKB=BKB, g=g, rho=rho, visc=visc, scale_factor=scale_factor, S_f=S_f,
                   sigma_limit=sigma_limit, R_b=R_b,
                   **kwargs)

    @property
    def OG(self):
        """
        distance from roll axis (cg) to still water level [m] (positive into water)
        """
        return self.draught - self.kg

    @property
    def Cb(self):
        return self.volume/(self.lpp*self.beam*self.draught)

    @property
    def A_mid(self):
        """
        Mid section area [m2]
        """
        return self.A0*self.beam*self.draught

    @property
    def Ho(self):
        """
        half breadth to draft ratio
        H0 = beam/(2*draught)
        """
        return self.beam/(2*self.draught)

    @property
    def R(self):
        """
        Bilge radius [m]
        """
        if hasattr(self, '_R'):
            return self._R
        else:
            return self.calculate_R_b()

    @R.setter
    def R(self, value):
        self._R = value

    @property
    def S_s(self):
        return self.C_s*self.T_s*self.B_s

    @property
    def A0(self):
        """
        mid section coefficient [-]
        """
        index = int(len(self.C_s)/2)
        return self.C_s[index]

    @property
    def B_s(self):
        # Sectional beam [m] (incl. both sides)
        B_s = np.array(self.sections['B_s'])
        mask=B_s==0
        B_s[mask]=0.000001*np.max(B_s) # Putin a very small value
        return B_s

    @property
    def T_s(self):
        # Sectional draught [m]
        T_s = np.array(self.sections['T_s'])
        mask = T_s == 0
        T_s[mask] = 0.000001*np.max(T_s)  # Putin a very small value
        return T_s

    @property
    def draught(self):
        # draught ship draught [m]
        return np.max(self.T_s)

    @property
    def C_s(self):
        # Sectional area coefficient: C_s = S_s/(B_s*T_s)
        C_s = np.array(self.sections['C_s'])

        ## Apply sigma limit:
        mask = C_s > self.sigma_limit
        C_s[mask] = self.sigma_limit

        return C_s

    @property
    def x_s(self):
        # Sectional x-coordinate from AP [m]
        return np.array(self.sections.index)

    @property
    def w_hat(self):
        return lambdas.omega_hat(beam=self.beam, g=self.g, omega0=self.w)

    @property
    def BD(self):
        return self.beam / self.draught

    @property
    def OGD(self):
        return self.OG / self.draught

    def verify_sections(self):

        mandatorys=['B_s', 'T_s', 'S_s']
        for mandatory in mandatorys:
            if not hasattr(self,mandatory):
                raise SectionsError('You need to specify sections for this ship')

    def B_hat(self,B):
        """
        Nondimensionalize the damping
        Parameters
        ----------
        B
            damping [Nm*rad/s]

        Returns
        -------
        B_hat
            Nondimensional damping coefficient
        """
        return lambdas.B_hat_lambda(B=B, Disp=self.volume, beam=self.beam, g=self.g, rho=self.rho)

    def calculate_sectional_lewis_coefficients(self):
        """
        Lewis form approximation' is obtained.
        Given the section's area, S_s, beam B_s and draught T_s, the constants a, a a_3 are uniquely defined
        by von Kerczek and Tuck:
        See code in: ikeda_speed.calculate_sectional_lewis

        Returns
        -------
        a, a_1, a_3, sigma, H0 : array_like
            sectional lewis coefficients.
        """
        self.verify_sections()
        self.a, self.a_1, self.a_3, self.sigma_s, self.H = ikeda_speed.calculate_sectional_lewis(B_s=self.B_s, T_s=self.T_s, S_s=self.S_s)
        return self.a, self.a_1, self.a_3, self.sigma_s, self.H

    def calculate(self, w=None, fi_a=None) -> pd.DataFrame:
        """
        Calculate all dampings

        Parameters
        ----------
        w : float or ndarray or None (if it has already been defined)
            frequancies (usually natural frequency)
        fi_a : float or ndarray or None (if it has already been defined)
            roll amplitude [rad]

        Returns
        -------
        pd.DataFrame
            DataFrame with ikeda damping
        """

        if w is None:
            if self.w is None:
                raise WInputError('Please specify w')
        else:
            self.w = w

        if fi_a is None:
            if self.fi_a is None:
                raise PhiAInputError('Please specify fi_a')
        else:
            self.fi_a = fi_a

        #if np.isscalar(self.lBK):
        #    if (len(self.fi_a) > 0):
        #        self.lBK *= np.ones(len(self.fi_a))
        #elif len(self.fi_a) != len(self.lBK):
        #        self.lBK = self.lBK[0]*np.ones(len(self.fi_a))
        #
        #if np.isscalar(self.bBK):
        #    if (len(self.fi_a) > 0):
        #        self.bBK *= np.ones(len(self.fi_a))
        #elif len(self.fi_a) != len(self.bBK):
        #        self.bBK = self.bBK[0] * np.ones(len(self.fi_a))

        output = pd.DataFrame()
        output['B_44_hat'] = self._component(output,self.calculate_B44())
        output['B_W0_hat'] = self._component(output,self.calculate_B_W0())
        output['B_W_hat'] = self._component(output,self.calculate_B_W())
        output['B_F_hat'] = self._component(output,self.calculate_B_F())
        output['B_E_hat'] = self._component(output,self.calculate_B_E())
        output['B_BK_hat'] = self._component(output,self.calculate_B_BK())
        output['B_L_hat'] =  self._component(output,self.calculate_B_L())
        output['Bw_div_Bw0'] = self._component(output,self.calculate_Bw_div_Bw0())

        return output

    @staticmethod
    def _component(df,component):

        if len(df) > len(component):
            values = np.tile(component,(len(df),1))
        else:
            values = component

        return values

    def calculate_B44(self):
        """
        Calculate total roll damping

        Returns
        -------
        B_44_hat : ndarray
            Total roll damping [-]
        """
        B_44_hat = (self.calculate_B_W() +
                self.calculate_B_F() +
                self.calculate_B_E() +
                self.calculate_B_L() +
                self.calculate_B_BK()
                )
        return B_44_hat

    def calculate_Bw_div_Bw0(self):
        """
        Calculate Wave damping speed correction

        Returns
        -------
        Bw_div_Bw0
            Bw_div_Bw0 = B_W/B_W0
        """
        self.Bw_div_Bw0 = ikeda_speed.B_W_speed_correction_factor_ikeda(w=self.w, V=self.V, d=self.draught, g=self.g)
        return array(self.Bw_div_Bw0)

    def calculate_B_W(self, Bw_div_Bw0_max=np.inf):
        """
        Calculate roll wave damping at speed

        Returns
        -------
        B_W_hat : ndarray
            Roll wave damping at speed [-]

        """
        B_W0_hat = self.calculate_B_W0()
        Bw_div_Bw0 = self.calculate_Bw_div_Bw0()
        B_W_hat = B_W0_hat*Bw_div_Bw0


        return array(B_W_hat)

    def calculate_B_W0(self):
        """
        Calculate roll wave damping at zero speed

        Returns
        -------
        B_W0_hat : ndarray
            Roll wave damping at zero speed [-]

        """

        w_hat = self.B_W0_hat.index
        B_W0_hat = np.interp(self.w_hat, w_hat, self.B_W0_hat)  # Zero speed wave damping [Nm*s/rad]

        return array(B_W0_hat)

    def calculate_B_F(self):
        """
        Calculate skin friction damping

        Returns
        -------
        B_F_hat : ndarray
            Skin friction damping [-]

        """

        B_F = ikeda_speed.frictional(w=self.w, fi_a=self.fi_a, V=self.V, B=self.beam, d=self.draught, OG=self.OG,
                                      ra=self.rho, Cb=self.Cb,L=self.lpp, visc=self.visc, Sf=self.S_f)
        return array(self.B_hat(B_F))

    def calculate_B_E0(self):
        """
        Calculate bilge eddy damping at zero speed

        Returns
        -------
        B_E0_hat : ndarray
            eddy damping [-]

        """
        a, a_1, a_3, sigma_s, H = self.calculate_sectional_lewis_coefficients()

        B_E0 = ikeda_speed.eddy(bwl=self.B_s, a_1=a_1, a_3=a_3, sigma=sigma_s, xs=self.x_s, H0=H, Ts=self.T_s, OG=self.OG,
                               R=self.R, wE=self.w, fi_a=self.fi_a)
        return array(self.B_hat(B_E0))

    def calculate_B_E(self):
        """
        Calculate bilge eddy damping at speed

        Returns
        -------
        B_E_hat : ndarray
            eddy damping [-]

        """
        B_E0_hat = self.calculate_B_E0()
        #factor=(0.04*self.w*self.lpp/self.V)**2
        factor = np.divide(0.04 * self.w * self.lpp, self.V, out=np.zeros_like(self.V), where=(self.V!=0))**2
        if isinstance(factor,float):
            factor = np.array([factor])

        factor_total = (factor)/(1+factor)
        mask=factor==0
        factor_total[mask]=1.0

        B_E_hat=B_E0_hat*factor_total
        return array(B_E_hat)

    def calculate_B_E0_sections(self):
        """
        Calculate bilge eddy damping for each section at zero speed

        Returns
        -------
        B_E0_s_hat : ndarray
            eddy damping for each sections [-]

        """
        a, a_1, a_3, sigma_s, H = self.calculate_sectional_lewis_coefficients()

        B_E0_s = ikeda_speed.eddy_sections(bwl=self.B_s, a_1=a_1, a_3=a_3, sigma=sigma_s, H0=H, Ts=self.T_s, OG=self.OG,
                               R=self.R, wE=self.w, fi_a=self.fi_a)

        return array(self.B_hat(B_E0_s))

    def calculate_R_b(self):
        """
        Calculate bilge radius with Ikedas empirical formula:
        Returns
        -------
        R_b : ndarray
            Bilge radius [m]

        """
        a, a_1, a_3, sigma_s, H = self.calculate_sectional_lewis_coefficients()

        mask=sigma_s>1
        sigma_s[mask]=0.99  # Do avoid negative value in sqrt
        mask=H<0
        R_b = 2*self.draught*np.sqrt(H*(sigma_s-1)/(np.pi-4))

        mask = (H>=1) & (R_b/self.draught>1)
        R_b[mask]=self.draught

        mask = (H < 1) & (R_b / self.draught > H)
        R_b[mask] = self.beam/2

        return R_b

    def calculate_B_L(self):
        """
        Calculate hull lift damping

        Returns
        -------
        B_L_hat : ndarray
            HUl lift damping [-]

        """
        B_L = ikeda_speed.hull_lift(V=self.V, d=self.draught, OG=self.OG, L=self.lpp, A=self.A_mid, ra=self.rho,
                                     B=self.beam)
        return array(self.B_hat(B_L))

    def calculate_B_BK(self):
        """
        Calculate bilge keel damping

        Returns
        -------
        B_BK_hat : ndarray
            Bilge keel damping [-]

        """

        if np.any(~(self.bBK==0) & (self.lBK==0)):
            raise BilgeKeelError('BKB is 0 but BKL is not!')
            return 0.0

        if isinstance(self.R, np.ndarray):
            index=int(len(self.R)/2)  # Somewhere in the middle of the ship
            R = self.R[index]
        else:
            R = self.R

        Bp44BK_N0, Bp44BK_H0, B44BK_L, B44BKW0 = ikeda_speed.bilge_keel(w=self.w, fi_a=self.fi_a, V=self.V, B=self.beam,
                                                                        d=self.draught, A=self.A_mid,
                                      bBK=self.bBK, R=R, g=self.g, OG=self.OG, Ho=self.Ho, ra=self.rho)

        B44BK_N0 = Bp44BK_N0*self.lBK
        B44BK_H0 = Bp44BK_H0*self.lBK
        B44_BK = B44BK_N0 + B44BK_H0 + B44BK_L

        mask = ((self.lBK == 0) | (pd.isnull(self.lBK)))

        if np.isscalar(mask):
            if mask:
                B44_BK = 0
        else:
            B44_BK[mask] = 0

        return array(self.B_hat(B44_BK))

def array(value):
    if isinstance(value,float):
        return np.array([value])
    else:
        return value

class IkedaR(Ikeda):
    """
    Same as Ikeda class but with mandatory bilge radius
    """

    def __init__(self, V: np.ndarray, w: np.ndarray, fi_a: float, B_W0_hat: pd.Series, beam: float, lpp: float,
                 kg: float, volume: float, sections: pd.DataFrame, BKL, BKB, R_b:float,
                 g=9.81, rho=1000.0, visc=1.15 * 10 ** -6, scale_factor=1.0, S_f=None, sigma_limit = 0.99, **kwargs):

        super().__init__(V=V, w=w, fi_a=fi_a, B_W0_hat=B_W0_hat, beam=beam, lpp=lpp,
                                    kg=kg, volume=volume, sections=sections, BKL=BKL, BKB=BKB,
                                    g=g, rho=rho, visc =visc, scale_factor=scale_factor, S_f=S_f,
                                    sigma_limit=sigma_limit,**kwargs)

        self._R = R_b

    @classmethod
    def load_scoresII(cls, V: np.ndarray, w: np.ndarray, fi_a: float, indata: pyscores2.indata.Indata,
                      output_file: pyscores2.output.OutputFile, BKL: float, BKB: float, kg, R_b:float, g=9.81, rho=1000.0,
                      visc=1.15 * 10 ** -6,
                      scale_factor=1.0, S_f=None, sigma_limit = 0.99,
                      **kwargs):

        estimator = super(cls,cls).load_scoresII(V=V, w=w, fi_a=fi_a, indata=indata, output_file=output_file, BKL=BKL, BKB=BKB, kg=kg, g=g, rho=rho, visc=visc,
                                        scale_factor=scale_factor, R_b=R_b, sigma_limit=sigma_limit, **kwargs)
        return estimator

class IkedaBeSimplified(Ikeda):
    """
       Same as Ikeda class but with eddy damping from Simplified Ikeda
    """

    def calculate_B_E0(self):
        """
        Calculate bilge eddy damping at zero speed

        Returns
        -------
        B_E0_hat : ndarray
            eddy damping [-]

        """
        a, a_1, a_3, sigma_s, H = self.calculate_sectional_lewis_coefficients()

        B_E0 = ikeda_speed.eddy(bwl=self.B_s, a_1=a_1, a_3=a_3, sigma=sigma_s, xs=self.x_s, H0=H, Ts=self.T_s, OG=self.OG,
                               R=self.R, d=self.draught, wE=self.w, fi_a=self.fi_a)

        BD = self.beam/self.draught
        CB = self.Cb
        CMID = self.A0
        OGD = self.OG/self.draught
        OMEGAHAT = self.w_hat
        PHI = np.rad2deg(self.fi_a)  # [deg] !
        B_E0_hat = simplified_ikeda.calculate_B_E(BD, CB, CMID, OGD, OMEGAHAT, PHI)

        return B_E0_hat



class IkedaCarlJohan(Ikeda):
    """
       Same as Ikeda class but with different kind of eddy damping
    """

    def calculate_B_E0(self):
        """
        Calculate bilge eddy damping at zero speed

        Returns
        -------
        B_E0_hat : ndarray
            eddy damping [-]

        """
        a, a_1, a_3, sigma_s, H = self.calculate_sectional_lewis_coefficients()

        B_E0 = ikeda_speed.eddy(bwl=self.B_s, a_1=a_1, a_3=a_3, sigma=sigma_s, xs=self.x_s, H0=H, Ts=self.T_s, OG=self.OG,
                               R=self.R, d=self.draught, wE=self.w, fi_a=self.fi_a)
        return self.B_hat(B_E0)

class IkedaBarge(Ikeda):
    """
       Same as Ikeda class but with barge eddy damping
    """

    def calculate_B_E0(self):
        """
        Calculate bilge eddy damping at zero speed

        Returns
        -------
        B_E0_hat : ndarray
            eddy damping [-]

        """
        B_E0 = ikeda_barge.eddy(rho=self.rho, lpp=self.lpp, d=self.draught, beam=self.beam, OG=self.OG, phi_a=self.fi_a, w=self.w)
        return self.B_hat(B_E0)

class IkedaS(Ikeda):
    """
        Same as Ikeda class but with mandatory wetted surface.
    """

    def __init__(self, V: np.ndarray, w: np.ndarray, fi_a: float, B_W0_hat: pd.Series, beam: float, lpp: float,
                 kg: float, volume: float, sections: pd.DataFrame, BKL, BKB, S_f:float,
                 g=9.81, rho=1000.0, visc=1.15 * 10 ** -6, scale_factor=1.0, **kwargs):

        super().__init__(V=V, w=w, fi_a=fi_a, B_W0_hat=B_W0_hat, beam=beam, lpp=lpp,
                 kg=kg, volume=volume, sections=sections, BKL=BKL, BKB=BKB,
                 g=g, rho=rho, visc =visc, scale_factor=scale_factor, S_f=S_f, **kwargs)

class IkedaCr(Ikeda):
    """
    Same as Ikeda class but with mandatory manual C_r
    """

#    def __init__(self, V: np.ndarray, w: np.ndarray, fi_a: float, B_W0_hat: pd.Series, beam: float, lpp: float,
#                 kg: float, volume: float, sections: pd.DataFrame, BKL, BKB, C_r,
#                 g=9.81, rho=1000.0, visc=1.15 * 10 ** -6, scale_factor=1.0, S_f=None, **kwargs):
#
#        super().__init__(V=V, w=w, fi_a=fi_a, B_W0_hat=B_W0_hat, beam=beam, lpp=lpp,
#                                    kg=kg, volume=volume, sections=sections, BKL=BKL, BKB=BKB,
#                                    g=g, rho=rho, visc =visc, scale_factor=scale_factor, S_f=S_f, **kwargs)
#
#        self.C_r = C_r
#
#    @classmethod
#    def load_scoresII(cls, V: np.ndarray, w: np.ndarray, fi_a: float, indata: pyscores2.indata.Indata,
#                      output_file: pyscores2.output.OutputFile, BKL: float, BKB: float, kg, C_r, g=9.81, rho=1000.0,
#                      visc=1.15 * 10 ** -6,
#                      scale_factor=1.0, S_f=None,
#                      **kwargs):
#
#        estimator = super(cls,cls).load_scoresII(V=V, w=w, fi_a=fi_a, indata=indata, output_file=output_file, BKL=BKL, BKB=BKB, kg=kg, g=g, rho=rho, visc=visc,
#                                        scale_factor=scale_factor, C_r=C_r, **kwargs)
#        return estimator

    def calculate_B_E0(self):
        """
        Calculate bilge eddy damping at zero speed using manually deined C_r

        Returns
        -------
        B_E0_hat : ndarray
            eddy damping [-]

        """
        Bp44E0s = self.eddy_sections()

        Bp44E0 = simps(y=Bp44E0s, x=self.x_s, axis=0)

        return self.B_hat(Bp44E0)

    def eddy_sections(self):

        # WE, CR = np.meshgrid(wE, Cr)
        Cr = np.array(self.C_r)
        wE = np.array(self.w)

        try:
            len_Cr = len(Cr)
        except TypeError:
            len_Cr = 1

        try:
            len_wE = len(wE)
        except TypeError:
            len_wE = 1

        WE = np.tile(wE, (len_Cr, 1))
        FI_a = np.tile(self.fi_a, (len_Cr, 1))

        Tss = np.tile(self.T_s, (len_wE, 1)).transpose()
        CR = np.tile(Cr, (len_wE, 1)).transpose()



        #Bp44E0s = 4 * self.rho * self.T_s**4 * self.w * self.fi_a * self.C_r / (3 * np.pi)

        Bp44E0s = 4 * self.rho * Tss ** 4 * WE * FI_a * CR / (3 * np.pi)

        return Bp44E0s


from scipy import signal

def butter_lowpass(cutoff, fs, order=5):
    nyq = 0.5 * fs
    normal_cutoff = cutoff / nyq
    b, a = signal.butter(order, normal_cutoff, btype='low', analog=False)
    return b, a

def lowpass_filter(data, cutoff, fs, order=5):

    if cutoff is None or order is None:
        # Disabled filter:
        return data

    b, a = butter_lowpass(cutoff, fs, order=order)
    #y = signal.lfilter(b, a, data)
    y = signal.filtfilt(b, a, data)

    return y


import pandas as pd
import numpy as np
import sympy as sp
import matplotlib.pyplot as plt
from rolldecayestimators.estimator import RollDecay
from rolldecayestimators.direct_estimator import DirectEstimator

from rolldecayestimators.simplified_ikeda import calculate_roll_damping
from rolldecayestimators import equations
from rolldecayestimators import symbols
from rolldecayestimators.substitute_dynamic_symbols import lambdify
from rolldecayestimators.sensitivity import variate_ship, plot_variation, calculate, calculate_variation, _plot_result


from scipy.optimize import curve_fit

class IkedaEstimatorFitError(ValueError): pass

class IkedaEstimator(DirectEstimator):

    eqs = [equations.zeta_equation,  # 0
           equations.omega0_equation_linear]  # 1
    functions_ikeda = [lambdify(sp.solve(eqs, symbols.A_44, symbols.zeta)[0][1]),
                       lambdify(sp.solve(equations.B44_equation, symbols.B_44)[0]),
                       ]

    def __init__(self, lpp:float, TA, TF, beam, BKL, BKB, A0, kg, Volume, gm, V, rho=1000, g=9.81, phi_max=8, omega0=None,
                 verify_input = True, limit_inputs=False, **kwargs):
        """
        Estimate a roll decay test using the Simplified Ikeda Method to predict roll damping.
        NOTE! This method is currently only valid for zero speed!

        Parameters
        ----------
        lpp
            Ship perpendicular length [m]
        TA
            Draught aft [m]
        TF
            Draught forward [m]
        beam
            Ship beam [m]
        BKL
            Bilge keel length [m]
        BKB
            Bilge keel height [m]
        A0
            Middship coefficient (A_m/(B*d) [-]
        kg
            Vertical centre of gravity [m]
        Volume
            Displacement of ship [m3]
        gm
            metacentric height [m]
        V
            ship speed [m/s]
        rho
            Density of water [kg/m3]
        g
            acceleration of gravity [m/s**2]
        phi_max
            max roll angle during test [deg]
        omega0
            Natural frequency of motion [rad/s], if None it will be calculated with fft of signal

        For more info see: "rolldecaysestimators/simplified_ikeda.py"
        """
        super().__init__(omega0=omega0)

        self.lpp=lpp
        self.TA=TA
        self.TF=TF
        self.beam=beam
        self.BKL=BKL
        self.BKB=BKB
        self.A0=A0
        self.kg=kg
        self.Volume=Volume
        self.V = V
        self.rho=rho
        self.g=g
        self.gm=gm
        self.phi_max=phi_max
        self.two_point_regression=True
        self.verify_input = verify_input
        self.limit_inputs = limit_inputs

    @property
    def zeta_lambda(self):
        return self.functions_ikeda[0]

    @property
    def B44_lambda(self):
        return self.functions_ikeda[1]

    #def simulate(self, t :np.ndarray, phi0 :float, phi1d0 :float,omega0:float, zeta:float)->pd.DataFrame:
    #    """
    #    Simulate a roll decay test using the quadratic method.
    #    :param t: time vector to be simulated [s]
    #    :param phi0: initial roll angle [rad]
    #    :param phi1d0: initial roll speed [rad/s]
    #    :param omega0: roll natural frequency[rad/s]
    #    :param zeta:linear roll damping [-]
    #    :return: pandas data frame with time series of 'phi' and 'phi1d'
    #    """
    #    parameters={
    #        'omega0':omega0,
    #        'zeta':zeta,
    #    }
    #    return self._simulate(t=t, phi0=phi0, phi1d0=phi1d0, parameters=parameters)

    def fit(self, X, y=None, **kwargs):
        self.X = X

        self.phi_max = np.rad2deg(self.X[self.phi_key].abs().max())  ## Initial roll angle in [deg]

        DRAFT=(self.TA + self.TF) / 2
        omega0=self.omega0

        if (self.lpp*self.beam*DRAFT > 0):
            CB = self.Volume / (self.lpp*self.beam*DRAFT)
        else:
            raise IkedaEstimatorFitError('lpp, beam or DRAFT is zero or nan!')

        self.ikeda_parameters = {

            'LPP' : self.lpp,
            'Beam' : self.beam,
            'DRAFT' : DRAFT,

            'PHI' : self.phi_max,
            'BKL' : self.BKL,
            'BKB' : self.BKB,
            'OMEGA' : omega0,
            'OG' : (-self.kg + DRAFT),
            'CB' : CB,
            'CMID' : self.A0,
            'V':self.V,
            'verify_input':self.verify_input,
            'limit_inputs':self.limit_inputs,

        }

        self.result=self.calculate()

        m = self.Volume * self.rho
        B_44 = self.B44_lambda(B_44_hat=self.result.B_44_HAT, Disp=self.Volume, beam=self.beam, g=self.g, rho=self.rho)
        zeta = self.zeta_lambda(B_1=B_44, GM=self.gm, g=self.g, m=m, omega0=omega0)
        self.parameters={
            'zeta':zeta,
            'omega0':omega0,
            'd':0,
        }

        self.is_fitted_ = True


    def calculate(self,**kwargs):

        B44HAT, BFHAT, BWHAT, BEHAT, BBKHAT, BLHAT = calculate_roll_damping(**self.ikeda_parameters, **kwargs)
        s = pd.Series()
        s['B_44_HAT'] = B44HAT
        s['B_F_HAT'] = BFHAT
        s['B_W_HAT'] = BWHAT
        s['B_E_HAT'] = BEHAT
        s['B_BK_HAT'] = BBKHAT
        s['B_L_HAT'] = BLHAT

        return s

    def result_for_database(self, score=True, **kwargs):

        s = super().result_for_database(score=score, **kwargs)
        s.update(self.result)

        return s

class IkedaQuadraticEstimator(IkedaEstimator):

    functions_ikeda = IkedaEstimator.functions_ikeda
    functions_ikeda.append(lambdify(sp.solve(equations.B_e_equation, symbols.B_e)[0]))       # 2
    functions_ikeda.append(lambdify(sp.solve(equations.zeta_B1_equation, symbols.zeta)[0]))  # 3
    functions_ikeda.append(lambdify(sp.solve(equations.d_B2_equation, symbols.d)[0]))        # 4

    @property
    def B_e_lambda(self):
        return self.functions_ikeda[2]

    @property
    def zeta_B1_lambda(self):
        return self.functions_ikeda[3]

    @property
    def d_B2_lambda(self):
        return self.functions_ikeda[4]

    def fit(self, X=None, y=None, **kwargs):
        self.X = X

        if not self.X is None:
            self.phi_max = np.rad2deg(self.X[self.phi_key].abs().max())  ## Initial roll angle in [deg]


        DRAFT=(self.TA + self.TF) / 2
        omega0=self.omega0

        if (self.lpp*self.beam*DRAFT > 0):
            CB = self.Volume / (self.lpp*self.beam*DRAFT)
        else:
            raise IkedaEstimatorFitError('lpp, beam or DRAFT is zero or nan!')

        self.ikeda_parameters = {

            'LPP' : self.lpp,
            'Beam' : self.beam,
            'DRAFT' : DRAFT,

            'PHI' : self.phi_max,
            'BKL' : self.BKL,
            'BKB' : self.BKB,
            'OMEGA' : omega0,
            'OG' : (-self.kg + DRAFT),
            'CB' : CB,
            'V'  : self.V,
            'CMID' : self.A0,
            'verify_input': self.verify_input,
            'limit_inputs': self.limit_inputs,
        }

        #self.result=self.calculate(**kwargs)
        self.result = {}

        self.result_variation=self.calculate_phi_a_variation()

        if self.two_point_regression:
            B_1_,B_2_ = self.calculate_two_point_regression(**kwargs)
            self.result['B_1'] = B_1 = B_1_['B_44']
            self.result['B_2'] = B_2 = B_2_['B_44']

            B_1_, B_2_ = self.fit_Bs()  # Not used...
        else:
            B_1, B_2 = self.fit_Bs()
            self.result['B_1'] = B_1
            self.result['B_2'] = B_2

        zeta, d = self.Bs_to_zeta_d(B_1=B_1, B_2=B_2)
        factor = 1.0  # Factor
        phi_a = np.abs(np.deg2rad(self.phi_max))/ factor  # [Radians]
        self.result['B_e'] = self.B_e_lambda(B_1=B_1, B_2=B_2, omega0=self.ikeda_parameters['OMEGA'],phi_a=phi_a)

        self.parameters={
            'zeta':zeta,
            'd':d,
            'omega0':omega0,
        }

        self.is_fitted_ = True

    def calculate_two_point_regression(self, **kwargs):
        # Two point regression:
        data = {
            'lpp': self.lpp,
            'beam': self.beam,
            'DRAFT': (self.TA + self.TF) / 2,
            'phi_max': self.phi_max,
            'BKL': self.BKL,
            'BKB': self.BKB,
            'omega0': self.ikeda_parameters['OMEGA'],
            'kg': self.kg,
            'CB': self.ikeda_parameters['CB'],
            'A0': self.A0,
            'V': self.V,
            'Volume': self.Volume,
        }
        row1 = pd.Series(data)
        row1.phi_max *= 0.5
        row2 = pd.Series(data)
        s1_hat = calculate(row1, verify_input=self.verify_input, limit_inputs=self.limit_inputs, **kwargs)
        s2_hat = calculate(row2, verify_input=self.verify_input, limit_inputs=self.limit_inputs, **kwargs)

        s1_hat = self.B44_lambda(B_44_hat=s1_hat, Disp=row1.Volume, beam=row1.beam, g=self.g, rho=self.rho)
        s2_hat = self.B44_lambda(B_44_hat=s2_hat, Disp=row2.Volume, beam=row2.beam, g=self.g, rho=self.rho)

        s1=pd.Series()
        s2=pd.Series()
        for key,value in s1_hat.items():
            new_key = key.replace('_hat','')
            s1[new_key]=s1_hat[key]
            s2[new_key] = s2_hat[key]

        x = np.deg2rad([row1.phi_max, row2.phi_max]) * 8 * row1.omega0 / (3 * np.pi)
        B_2 = (s2 - s1) / (x[1] - x[0])
        B_1 = s1 - B_2 * x[0]

        # Save all of the component as one linear term: _1 and a quadratic term: _2
        for key in s1.index:
            new_name_1 = '%s_1' % key
            self.result[new_name_1] = s1[key]

            new_name_2 = '%s_2' % key
            self.result[new_name_2] = s2[key]

        return B_1,B_2

    def calculate_phi_a_variation(self):

        data={
            'lpp':self.lpp,
            'beam' : self.beam,
            'DRAFT' : (self.TA+self.TF)/2,
            'phi_max' : self.phi_max,
            'BKL' :  self.BKL,
            'BKB':  self.BKB,
            'omega0' : self.ikeda_parameters['OMEGA'],
            'kg' :self.kg,
            'CB':self.ikeda_parameters['CB'],
            'A0' : self.A0,
            'V' : self.V,
            'Volume':self.Volume,
        }
        self.ship = ship = pd.Series(data)

        N = 40
        changes = np.linspace(1, 0.0001, N)
        df_variation = variate_ship(ship=ship, key='phi_max', changes=changes)
        result = calculate_variation(df=df_variation, limit_inputs=self.limit_inputs, verify_input=self.verify_input)
        df_variation['g'] = 9.81
        df_variation['rho'] = 1000
        result = pd.concat((result, df_variation), axis=1)

        result['B_44'] = self.B44_lambda(B_44_hat=result.B_44_hat, Disp=ship.Volume, beam=ship.beam, g=result.g, rho=result.rho)
        result.dropna(inplace=True)
        return result

    def fit_Bs(self):

        def fit(df, B_1, B_2):
            omega0 = df['omega0']
            phi_a = np.deg2rad(df['phi_max'])  # Deg or rad (Radians gave better results actually)???
            #phi_a = df['phi_max']  # Deg or rad???
            return self.B_e_lambda(B_1, B_2, omega0, phi_a)

        coeffs, _ = curve_fit(f=fit, xdata=self.result_variation, ydata=self.result_variation['B_44'])
        B_1 = coeffs[0]
        B_2 = coeffs[1]
        self.result_variation['B_44_fit'] = fit(self.result_variation, *coeffs)
        return B_1,B_2

    def Bs_to_zeta_d(self, B_1, B_2):
        m = self.Volume*self.rho
        zeta = self.zeta_B1_lambda(B_1=B_1, GM=self.gm, g=self.g, m=m, omega0=self.ikeda_parameters['OMEGA'])
        d = self.d_B2_lambda(B_2=B_2, GM=self.gm, g=self.g, m=m, omega0=self.ikeda_parameters['OMEGA'])
        return zeta,d

    def plot_variation(self,ax=None):

        if ax is None:
            fig,ax=plt.subplots()

        self.result_variation.plot(y = ['B_44_hat'], ax=ax)

    def plot_B_fit(self,ax=None):

        if ax is None:
            fig,ax=plt.subplots()

        self.result_variation.plot(y='B_44', ax=ax)
        self.result_variation.plot(y='B_44_fit', ax=ax, style='--')

    @classmethod
    def load(cls, data: {}, X=None):
        """
        Load data and parameters from an existing fitted estimator

        Parameters
        ----------
        data : dict
            Dict containing data for this estimator such as parameters
        X : pd.DataFrame
            DataFrame containing the measurement that this estimator fits (optional).
        Returns
        -------
        estimator
            Loaded with parameters from data and maybe also a loaded measurement X
        """
        estimator = cls(**data)
        estimator.load_data(data=data)
        estimator.load_X(X=X)
        return estimator





import numpy as np
from numpy import pi, abs, tanh, cos, sin, exp, sqrt, arccos

def calculate_rmax(M, a_1, a_3, psi):
    r_maxs = M * sqrt(
        ((1 + a_1) * sin(psi) - a_3 * sin(3 * psi)) ** 2 + ((1 - a_1) * cos(psi) + a_3 * cos(3 * psi)) ** 2)
    return r_maxs

def eddy_sections(bwl:np.ndarray, a_1:np.ndarray, a_3:np.ndarray, sigma:np.ndarray, H0:np.ndarray, Ts:np.ndarray,
                  OG:float, R_b:float, wE:float, fi_a:float, rho=1000.0):
    """
    Calculation of eddy damping according to Ikeda.
    This implementation is a translation from:
    Ikeda, Y., 1978. On eddy making component of roll damping force on naked hull. University of Osaka Prefacture,
    Department of Naval Architecture, Japan, Report No. 00403,
    Published in: Journal of Society of Naval Architects of Japan, Volume 142.

    Parameters
    ----------
    bwl
        sectional beam water line [m]
    a_1
        sectional lewis coefficients
    a_3
        sectional lewis coefficients
    sigma
        sectional coefficient
    H0
        sectional coefficient
    Ts
        sectional draft [m]
    OG
        vertical distance water line to cg [m]
    R_b
        bilge radius [m]
    rho
        water density [kg/m3]
    wE
        roll requency [rad/s]
    fi_a
        roll amplitude [rad]

    Returns
    -------
    BE0s_hat
        Eddy damping per unit length for the sections at zero speed.
    """

    d = Ts

    B = bwl  # Probably
    M = B / (1 + a_1 + a_3)

    f_1 = 1/2*(1+tanh(20*(sigma-0.7)))
    f_2 = 1/2*(1-cos(pi*sigma))-1.5*(1-exp(-5*(1-sigma)))*sin(pi*sigma)**2
    H0_prim = H0*d/(d-OG)
    sigma_prim = (sigma*d-OG)/(d-OG)
    f_3 = 1 + 4*exp(-1.65*10**5*(1-sigma)**2)

    psi_1 = 0.0
    psi_2 = 1 / 2 * arccos(a_1 * (1 + a_3) / (4 * a_3))

    r_max_1 = calculate_rmax(M=M, a_1=a_1, a_3=a_3, psi=psi_1)
    r_max_2 = calculate_rmax(M=M, a_1=a_1, a_3=a_3, psi=psi_2)

    mask = r_max_1 >= r_max_2
    psi = psi_2
    psi[mask] = psi_1

    ## A
    cs = [-2 * a_3, a_1*(1 - a_3), ((6 - 3*a_1)*a_3**2 + (a_1**2 - 3*a_1)*a_3 + a_1**2)]
    ps = [5*psi, 3*psi, psi]
    A_1=0
    for c,p in zip(cs,ps):
        A_1+=c*cos(p)

    ## B
    B_1 = 0
    cs = [-2 * a_3, a_1 * (1 - a_3), ((6 + 3 * a_1)*a_3**2 + (3*a_1 + a_1**2)*a_3 + a_1**2)]
    for c, p in zip(cs, ps):
        B_1 += c * sin(p)


    r_max = calculate_rmax(M=M, a_1=a_1, a_3=a_3, psi=psi)

    H = 1 + a_1**2 + 9*a_3**2 + 2*a_1*(1-3*a_3)*cos(2*psi) - 6*a_3*cos(4*psi)
    V_max_div_phi1d = f_3*(r_max + 2*M/H*sqrt(A_1**2+B_1**2))
    gamma = sqrt(pi)/(2*d*(1-OG/d)*sqrt(H0_prim*sigma_prim))*V_max_div_phi1d

    C_p = 1/2*(0.87*exp(-gamma)-4*exp(-0.187*gamma)+3)


    R_b = calculate_R_b(beam=bwl, draught=Ts, H0=H0, sigma=sigma)

    M_re = 1/2*rho*r_max**2*d**2*C_p*((1-f_1*R_b/d)*(1-OG/d-f_1*R_b/d) + f_2*(H0-f_1*R_b/d)**2)
    C_r = M_re / (1/2*rho*d**4)
    BE0s = 4/(3*pi)*d**4*wE*fi_a*C_r  # Rewritten without hat
    return BE0s

def calculate_R_b(beam, draught, H0, sigma):
    """
    Calculate bilge radius with Ikedas empirical formula:
    Returns
    -------
    R_b : ndarray
        Bilge radius [m]

    """

    mask=sigma>1

    sigma[mask]=0.99  # Do avoid negative value in sqrt
    mask=H0<0
    R_b = 2*draught*np.sqrt(H0*(sigma-1)/(pi-4))

    mask = (H0>=1) & (R_b/draught>1)
    R_b[mask]=draught

    mask = (H0 < 1) & (R_b / draught > H0)
    R_b[mask] = beam/2

    return R_b


import os
from .direct_estimator import DirectEstimator
from .transformers import CutTransformer

from ._version import __version__

__all__ = ['DirectEstimator', '__version__']

path = os.path.dirname(__file__)
